from datetime import datetime

import dask
import dask.array as da
import h5py
import numpy as np
import pint_xarray  # noqa: F401
import xarray as xr

from peaks.core.fileIO.base_arpes_data_classes.base_arpes_data_class import (
    BaseARPESDataLoader,
    ureg,
)
from peaks.core.fileIO.loc_registry import register_loader


@register_loader
class BaseFeSuMaDataLoader(BaseARPESDataLoader):
    # Define class variables
    _loc_name = "FeSuMa"

    _desired_dim_order = ["detector_x", "eV", "detector_y"]

    @classmethod
    def _load_data(cls, fpath, lazy):
        class FeSuMaImageCube:
            def __init__(self, hdf5_file):
                self.hdf5_file = hdf5_file

                with h5py.File(hdf5_file, "r") as f:
                    # Find all groups that start with '0'
                    self.scans_list = [key for key in f.keys() if key.startswith("0")]

                    # Get the shape and dtype from the first image
                    image0 = f[f"{self.scans_list[0]}/analysisImage"]
                    self.image_shape = image0.shape
                    self.dtype = image0.dtype
                    start = image0.attrs.get("DimOffsetX")
                    delta = image0.attrs.get("DimDeltaX")
                    self.detector_x = np.arange(
                        start, start + delta * image0.shape[0], delta
                    )
                    self.detector_x_unit = image0.attrs.get("ScaleUnitsX")
                    if isinstance(self.detector_x_unit, bytes):
                        self.detector_x_unit = self.detector_x_unit.decode()
                    start = image0.attrs.get("DimOffsetY")
                    delta = image0.attrs.get("DimDeltaY")
                    self.detector_y = np.arange(
                        start, start + delta * image0.shape[1], delta
                    )
                    self.detector_y_unit = image0.attrs.get("ScaleUnitsY")
                    if isinstance(self.detector_y_unit, bytes):
                        self.detector_y_unit = self.detector_y_unit.decode()

                    # Parse Acquisition metadata as integers
                    self.steps = np.array(
                        [
                            int(f[f"{scan}/AcquisitionCurrentStep"][0])
                            for scan in self.scans_list
                        ]
                    )
                    self.sweeps = np.array(
                        [
                            int(f[f"{scan}/AcquisitionCurrentSweep"][0])
                            for scan in self.scans_list
                        ]
                    )

                    # Get acquisition co-ordinate
                    self.acquisition_coord = f["AcquisitionCoordinateNice"][0]
                    if isinstance(self.acquisition_coord, bytes):
                        self.acquisition_coord = self.acquisition_coord.decode()

                    # Get the KE scaling
                    try:
                        # If a KE scan, should list the relevant KE values in the common metadata
                        # Don't trust AcquisitionEkinStop - this sometimes includes an extra step
                        self.eV = np.linspace(
                            float(f["AcquisitionEkinStart"][0]),
                            float(f["AcquisitionEkinStart"][0])
                            + (f["AcquisitionEkinStep"][0] * (len(set(self.steps)) - 1)),
                            len(set(self.steps)),
                        )
                    except KeyError:
                        # Get this from the first scan analyser voltages
                        self.eV = -f[f"{self.scans_list[0]}/AnalyzerUserSetVoltages"][3]

                    # Get delay positions if required
                    if self.acquisition_coord == "Delay Stage":
                        self.delay_pos = np.linspace(
                            float(f["AcquisitionDelayStart"][0]),
                            float(f["AcquisitionDelayStart"][0])
                            + (
                                f["AcquisitionDelayStep"][0] * (len(set(self.steps)) - 1)
                            ),
                            len(set(self.steps)),
                        )

            def __getitem__(self, idx):
                # Lazily load the slice when accessed
                scan_path = f"{self.scans_list[idx]}/analysisImage"
                with h5py.File(self.hdf5_file, "r") as f:
                    data = f[scan_path][:]
                return data

            def __len__(self):
                return len(self.scans_list)

            def shape(self):
                return (len(self.scans_list), *self.image_shape)

        # Instantiate the lazy loader
        lazy_cube = FeSuMaImageCube(fpath)

        # Create a list of Dask arrays from delayed objects
        lazy_slices = [
            da.from_delayed(
                dask.delayed(lazy_cube[i]),
                shape=lazy_cube.image_shape,
                dtype=lazy_cube.dtype,
            )
            for i in range(len(lazy_cube))
        ]

        # Stack the Dask arrays along a new dimension to form the 3D cube
        lazy_data = da.stack(lazy_slices, axis=0)

        # Create the xarray.DataArray with Acquisition metadata as dimensions
        data_array = xr.DataArray(
            lazy_data,
            dims=[
                "steps",
                "detector_x",
                "detector_y",
            ],  # Specifying dimensions for 3D data
            coords={
                "detector_x": lazy_cube.detector_x,
                "detector_y": lazy_cube.detector_y,
                "steps": lazy_cube.steps,  # Coordinates for the "steps" dimension
                "sweep": (
                    "steps",
                    lazy_cube.sweeps,
                ),  # Associating "sweep" with "steps"
            },
            name="spectrum",
        )

        if not lazy:
            data_array = data_array.compute()

        data_array = data_array.groupby("steps").mean()

        if lazy_cube.acquisition_coord == "Kinetic Energy":
            data_array = data_array.assign_coords({"eV": ("steps", lazy_cube.eV)})
            data_array = data_array.swap_dims({"steps": "eV"}).drop_vars("steps")
            scan_dim = "eV"
            scan_units = "eV"
        elif lazy_cube.acquisition_coord == "Delay Stage":
            data_array = data_array.assign_coords(
                {"delay_pos": ("steps", lazy_cube.delay_pos)}
            )
            data_array = data_array.swap_dims({"steps": "delay_pos"}).drop_vars("steps")
            scan_dim = "delay_pos"
            scan_units = "mm"

        return {
            "spectrum": data_array.data,
            "dims": data_array.dims,
            "coords": {dim: coord.data for dim, coord in data_array.coords.items()},
            "units": {
                scan_dim: scan_units,
                "detector_x": lazy_cube.detector_x_unit,
                "detector_y": lazy_cube.detector_y_unit,
                "spectrum": "counts",
            },
        }

    @classmethod
    def _load_metadata(cls, fpath):
        with h5py.File(fpath, "r") as f:
            image0 = f[
                f"{next(key for key in f.keys() if key.startswith('0'))}/analysisImage"
            ]
            dwell_unit = image0.attrs.get("ExposureTime_UNIT")

            # Parse KE
            # Get the KE scaling
            try:
                # If a KE scan, should list the relevant KE values in the common metadata
                # Don't trust AcquisitionEkinStop - this sometimes includes an extra step
                eV = [
                    float(f["AcquisitionEkinStart"][0]),
                    float(f["AcquisitionEkinStop"][0]),
                ] * ureg("eV")
                eV_step = f["AcquisitionEkinStep"][0] * ureg("eV")
            except KeyError:
                # Get this from the first scan analyser voltages
                eV = -f[
                    f"{next(key for key in f.keys() if key.startswith('0'))}/AnalyzerUserSetVoltages"
                ][3] * ureg("eV")
                eV_step = None

            if isinstance(dwell_unit, bytes):
                dwell_unit = dwell_unit.decode()
            metadata = {
                "analyser_sweeps": int(f["AcquisitionNumberOfSweeps"][0]),
                "analyser_dwell": image0.attrs.get("ExposureTime")[()]
                * image0.attrs.get("AccumulationCount")[()]
                * ureg(dwell_unit),
                "analyser_eV": eV,
                "analyser_step_size": eV_step,
                "analyser_eV_type": "kinetic energy cutoff",
                "timestamp": datetime.strptime(
                    f["AcquisitionStartTimeNice"][0].decode(), "%Y-%m-%dT%H:%M:%S.%fZ"
                ).strftime("%Y-%m-%d %H:%M:%S"),
            }

        return metadata



from typing import Optional, Union

import numpy as np
import pint
import pint_xarray
from pydantic import BaseModel, ConfigDict
from pydantic_core import core_schema

# Define the appropriate unit registry
ureg = pint_xarray.unit_registry


# Class to handle storing and validating pint Quantities in pydantic model
class Quantity(pint.Quantity):
    @classmethod
    def __get_pydantic_core_schema__(cls, source_type, handler):
        # Use with_info_plain_validator_function to handle both value and info
        return core_schema.with_info_plain_validator_function(cls.validate)

    @classmethod
    def validate(cls, v, info):
        if isinstance(v, pint.Quantity):
            return v
        elif isinstance(v, dict):
            value = v.get("value")
            units = v.get("units")
            if value is not None and units is not None:
                if isinstance(value, list):  # Handle list serialization of ndarray
                    value = np.array(value)
                return value * ureg(units)
            else:
                raise ValueError(
                    'Invalid quantity dictionary. Must have "value" and "units".'
                )
        elif (
            isinstance(v, tuple)
            and len(v) == 2
            and isinstance(v[1], str)
            and isinstance(v[0], (int, float, list, np.ndarray))
        ):
            return v[0] * ureg(v[1])
        elif isinstance(v, (int, float)):
            return v * ureg("")
        elif isinstance(v, np.ndarray):
            return v * ureg("")
        else:
            raise TypeError(f"Invalid type for Quantity: {type(v)}")

    def __repr__(self):
        return f"{self.magnitude} {self.units}"


# Handle serialising
def _quantity_encoder(quantity: pint.Quantity):
    if isinstance(quantity.magnitude, np.ndarray):
        # Convert ndarray to list for JSON serialization
        return {"value": quantity.magnitude.tolist(), "units": str(quantity.units)}
    return {"value": quantity.magnitude, "units": str(quantity.units)}


# Base class for passing a pint Quantity
class BaseMetadataModel(BaseModel):
    """Generalized model to store metadata, allowing serialising pint.Quantity objects."""

    model_config = ConfigDict(
        json_encoders={pint.Quantity: _quantity_encoder}, validate_assignment=True
    )


class BaseScanMetadataModel(BaseMetadataModel):
    """Model to store basic scan identifier metadata."""

    name: str
    filepath: str
    loc: str
    timestamp: str
    scan_command: Optional[str] = None


class NamedAxisMetadataModel(BaseMetadataModel):
    """Based model to store metadata for a single named axis.

    Attributes
    ----------
    local_name : Optional[str]
        The name of the axis.
    value : Optional[Union[Quantity, str, None]]
        The value of the axis.

    Methods
    -------
    set(value)
        Set the value of the axis.
    get()
        Get the value of the axis.

    local_name : Optional[str]
        The name of the axis.
    value : Optional[Union[Quantity, str, None]]
        The value of the axis.

    """

    local_name: Optional[str] = None
    value: Optional[Union[str, Quantity]] = None


# Define the manipulator metadata models
class AxisMetadataModelWithReference(NamedAxisMetadataModel):
    """Base model to store metadata for a single axis (e.g. manipulator axis) where a reference value is needed.

    Attributes
    ----------
    name : Optional[str]
        The local name of the axis on the actual system manipulator.
    value : Optional[Union[Quantity, str, None]]
        The value of the axis. If an array (e.g. for movement during the scan), this should return a string that
        describes the axis movement in the form x0:x_step:x1.
    reference_value : Optional[Union[Quantity, None]]
        The reference value of the axis. Supplied as None on initial load, but can be used later for keeping track
        e.g. of normal emission values.

    Methods
    -------
    In addition to the methods of NamedAxisMetadataModel, this class also has the following methods:
    set_reference(value)
        Set the reference value of the axis.
    """

    reference_value: Optional[Quantity] = None


# Define the temperature metadata models
class TemperatureMetadataModel(BaseMetadataModel):
    """Model to store temperature metadata."""

    sample: Optional[Union[str, Quantity]] = None
    cryostat: Optional[Union[str, Quantity]] = None
    shield: Optional[Union[str, Quantity]] = None
    setpoint: Optional[Union[str, Quantity]] = None


# Define the photon metadata models
class PhotonMetadataModel(BaseMetadataModel):
    """Model to store photon-linked metadata."""

    hv: Optional[Union[str, Quantity]] = None
    polarisation: Optional[Union[str, int, float]] = None
    exit_slit: Optional[Union[str, Quantity]] = None


class PumpPhotonMetadataModel(BaseMetadataModel):
    """Model to store pump beam metadata for pump-probe experiments."""

    hv: Optional[Union[str, Quantity]] = None
    polarisation: Optional[Union[str, int, float]] = None
    power: Optional[Union[str, Quantity]] = None
    delay: Optional[Union[str, Quantity]] = None
    t0_position: Optional[Union[str, Quantity]] = None


class ARPESSlitMetadataModel(BaseMetadataModel):
    """Model to store slit width metadata."""

    width: Optional[Union[str, Quantity]] = None
    identifier: Optional[str] = None


class ARPESAnalyserMetadataModel(BaseMetadataModel):
    """Model to store core analyser-linked metadata."""

    model: Optional[str] = None
    slit: ARPESSlitMetadataModel = ARPESSlitMetadataModel()


class ARPESScanMetadataModel(BaseMetadataModel):
    """Model to store scan metadata."""

    eV: Optional[Union[str, Quantity]] = None
    step_size: Optional[Union[str, Quantity]] = None
    PE: Optional[Union[str, Quantity]] = None
    sweeps: Optional[int] = None
    dwell: Optional[Union[str, Quantity]] = None
    lens_mode: Optional[str] = None
    acquisition_mode: Optional[str] = None
    eV_type: Optional[str] = None


class ARPESAnalyserAnglesMetadataModel(BaseMetadataModel):
    """Model to store analyser angles metadata.
    This is a general definiton which allows distinguishing analyser type (slit parallel or perp to slit)
    as well as allowing for the general case where the analyser can move."""

    polar: Optional[Union[str, Quantity]] = None
    tilt: Optional[Union[str, Quantity]] = None
    azi: Optional[Union[str, Quantity]] = None


class ARPESDeflectorMetadataModel(BaseMetadataModel):
    """Model to store deflector metadata."""

    parallel: NamedAxisMetadataModel = NamedAxisMetadataModel()
    perp: NamedAxisMetadataModel = NamedAxisMetadataModel()


class ARPESMetadataModel(BaseMetadataModel):
    """Model to store ARPES metadata."""

    analyser: ARPESAnalyserMetadataModel = ARPESAnalyserMetadataModel()
    scan: ARPESScanMetadataModel = ARPESScanMetadataModel()
    angles: ARPESAnalyserAnglesMetadataModel = ARPESAnalyserAnglesMetadataModel()
    deflector: ARPESDeflectorMetadataModel = ARPESDeflectorMetadataModel()


class ARPESCalibrationModel(BaseMetadataModel):
    EF_correction: Optional[
        Union[
            float,
            dict,
            int,
        ]
    ] = None
    V0: Optional[Union[str, Quantity]] = None



import importlib
import json
import re
from typing import Optional

import pint_xarray  # noqa: F401
import xarray as xr
from pydantic import create_model

from peaks.core.fileIO.base_data_classes.base_data_class import BaseDataLoader
from peaks.core.fileIO.loc_registry import register_loader
from peaks.core.metadata.base_metadata_models import AxisMetadataModelWithReference
from peaks.core.options import opts


# Classes for data loaders
@register_loader
class NetCDFLoader(BaseDataLoader):
    """Class for re-opening NetCDF files saved from :class:`peaks.core.fileIO.data_saving.save`.

    Notes
    -----
    If this is a NetCDF file, it should be a single DataArray or DataSet
    """

    _loc_name = "NetCDF"

    @classmethod
    def _load(cls, fpath, lazy, metadata, quiet):
        """Load the data from a NetCDF file.

        Parameters
        ----------
        fpath : str
            Path to the file to be loaded.
        lazy : bool
            Whether to load the data lazily.
        metadata : bool
            Whether to attempt to parse the metadata into full `peaks` format.
        quiet : bool
            Whether to suppress print output.

        Returns
        -------
        data : xarray.DataArray or xarray.DataSet
            The loaded data.
        """

        # Open NetCDF file as xarray.DataArray or xarray.Dataset
        try:
            data = xr.open_dataarray(fpath)
        except ValueError:
            data = xr.open_dataset(fpath)

        # Parse the metadata if requested
        if metadata:
            cls._parse_metadata(data)

        # Actually load the data
        if (lazy is False) or (lazy is None and data.nbytes > opts.FileIO.lazy_size):
            data = data.compute()

        # Quantify the data if it has units
        try:
            data = data.pint.quantify()
        except AttributeError:
            pass

        cls._add_load_history(data, fpath)

        return data

    @classmethod
    def _parse_metadata(cls, data):
        """Parse the metadata from the loaded data, returning to `peaks` format."""

        if data.attrs.get("metadata_models"):
            # Try to parse the loc
            loc = None
            loc_pattern = r'"loc":"(.*?)"'
            if data.attrs.get("_scan"):
                match = re.search(loc_pattern, data.attrs.get("_scan"))
                if match:
                    loc = match.group(1)

            metadata_models = json.loads(data.attrs.pop("metadata_models"))
            for attr_name, attr in data.attrs.items():
                model = metadata_models.get(attr_name)
                if model == "json":
                    data.attrs[attr_name] = json.loads(attr)
                elif model and "ManipulatorMetadataModel" in model:
                    # Need to handle this as a special case as the ManipulatorMetadataModel
                    # is created dynamically in loaders

                    # Get the axes for the original loader
                    manipulator_axes = cls.get_loader(loc)._manipulator_axes

                    # Rebuild manipulator metadata model
                    fields = {
                        axis: (Optional[AxisMetadataModelWithReference], None)
                        for axis in manipulator_axes
                    }
                    ManipulatorMetadataModel = create_model(
                        "ManipulatorMetadataModel", **fields
                    )
                    data.attrs[attr_name] = ManipulatorMetadataModel.parse_raw(attr)
                elif model:
                    model_class = cls._get_metadata_model(model)
                    data.attrs[attr_name] = model_class.parse_raw(attr)

    @classmethod
    def _get_metadata_model(cls, class_path):
        """Dynamically load the relevant metadata class from the fully qualified class name

        Parameters
        ----------
        class_path : str
            Fully qualified class name

        Returns
        -------
        class
            The class object
        """

        # Split the class_path into module and class parts
        module_name, class_name = class_path.rsplit(".", 1)

        # Dynamically import the module
        module = importlib.import_module(module_name)

        # Retrieve the class from the module
        return getattr(module, class_name)



"""Underlying custom fitting models or extensions for lmfit."""

import re

import lmfit.models as lm_models
import numpy as np
import xarray as xr
from lmfit import CompositeModel, Model
from scipy.ndimage import gaussian_filter1d

from peaks.core.fitting.fit_functions import _fermi_function, _linear_dos_fermi


def create_xarray_compatible_lmfit_model(model):
    """Dynamic class factory function to generate modified versions of lmfit models which accept 1D
    :class:`xarray.DataArray`'s as input for the guess parameters function.

    Parameters
    ----------
    model : lmfit.Model
        The :class:`lmfit` model.

    Returns
    -------
    WrappedModel : lmfit.Model
        The modified lmfit model which accepts 1D xarray.DataArray's as input for the guess parameters function.
    """

    class WrappedModel(model):
        def guess(self, data, **kws):
            if not isinstance(data, xr.DataArray):
                raise TypeError(
                    "This is a modified lmfit model which expects the data to be supplied as a 1D xarray.DataArray. "
                    "To pass numpy arrays, use the original lmfit model."
                )
            if data.ndim != 1:
                raise ValueError(
                    "Supplied xr.DataArray should be one dimensional, with the included dimension representing the "
                    "independent variable for the fit."
                )
            return super().guess(
                data=data.pint.dequantify().data, x=data[data.dims[0]].data, **kws
            )

        # Modify the docstring
        _original_guess_docstring = model.guess.__doc__
        _modified_guess_docstring = (
            f"Modified version of :class:`lmfit.model.Model.guess` method to accept a 1D "
            f"xarray.DataArray instead of seperate numpy arrays for data and independent "
            f"variable.\n{_original_guess_docstring}."
        )
        _new_guess_parameters = """
            Parameters
            ----------
            data : xarray.DataArray
                The data to guess fit parameters for. Should be a 1D DataArray where the dimension coresponds to the
                independent variable.
            **kws : optional
                Additional keyword arguments, passed to model function.
            """

        # Replace text between "Parameters" and "Returns"
        pattern = re.compile(
            r"(Parameters\s*[-]+\s*)(.*?)(\s*Returns\s*[-]+\s*)", re.DOTALL
        )
        _modified_guess_docstring = re.sub(
            pattern, r"\1" + _new_guess_parameters + r"\3", _modified_guess_docstring
        )
        guess.__doc__ = _modified_guess_docstring

    WrappedModel.__doc__ = (
        f"Modified version of :class:`lmfit.models.{model.__name__}` model with some key methods "
        f"modified to accept xarray.DataArray as inputs.\n{model.__doc__}"
    )
    return WrappedModel


# Get wrapped versions of all standard lmfit models
for model in lm_models.lmfit_models.values():
    model_name = model.__name__
    wrapped_model = create_xarray_compatible_lmfit_model(model)
    globals()[model_name] = wrapped_model


class GaussianConvolvedFitModel(create_xarray_compatible_lmfit_model(CompositeModel)):
    """Create a Gaussian convolved model for fitting, useful for e.g. including experimental resolution.

    Parameters
    ----------
    model : lmfit.Model or lmfit.CompositeModel
        The base or composite model to convolve.

    Examples
    --------
    Example usage is as follows::

        import peaks as pks
        import lmfit

        # Create a model - in this case a single Lorentzian peak
        model = lmfit.models.LorentzianModel()

        # Create a convolved model
        convolved_model = pks.GaussianConvolvedFitModel(model)
    """

    def __init__(self, model):
        def _gauss_conv(x, sigma_conv):
            """Convert sigma_conv from eV units to pixels."""
            return sigma_conv / (abs(x[-1] - x[0]) / len(x))

        def _convolve_gauss(model, sigma_conv_pxl):
            """Apply Gaussian convolution."""
            return gaussian_filter1d(model, sigma_conv_pxl)

        gauss_model = Model(_gauss_conv)
        gauss_model.set_param_hint("sigma_conv", min=0, value=0.001, vary=False)
        super().__init__(model, gauss_model, _convolve_gauss)


class FermiFunctionModel(create_xarray_compatible_lmfit_model(Model)):
    """lmfit compatible model for the Fermi function."""

    def __init__(self, *args, **kwargs):
        super().__init__(_fermi_function, *args, **kwargs)
        self.set_param_hint("EF", min=-np.inf, max=np.inf)
        self.set_param_hint("T", value=10, min=0, vary=False)

    def guess(self, data, **kws):
        """Guess the parameters of the model."""
        params = self.make_params()
        params[f"{self._prefix}EF"].set(value=data.estimate_EF())
        params[f"{self._prefix}T"].set(value=data.attrs.get("temp_sample", 10))
        return params


class LinearDosFermiModel(GaussianConvolvedFitModel):
    """Obtain a lmfit model for fitting typical poly-Au Fermi edge data. Includes: linear bg above and below E_F (dos_),
    Fermi cutoff (fermi_), linear background above E_F accounting for e.g. inhomogeneous detector efficiency (bg_) and
    Gaussian broadening for experimental energy resolution (conv_sigma).

    Attributes
    ----------
    model : lmfit.CompositeModel
        The composite model for fitting.
    """

    def __init__(self, prefix="", *args, **kwargs):
        self.base_model_prefix = prefix
        base_model = Model(_linear_dos_fermi, *args, prefix=prefix, **kwargs)
        base_model.set_param_hint("EF", min=-np.inf, max=np.inf)
        base_model.set_param_hint("T", value=10, min=0, vary=False)
        base_model.set_param_hint("bg_slope", value=0)
        base_model.set_param_hint("bg_intercept", value=0)
        base_model.set_param_hint("dos_slope", value=0)
        super().__init__(base_model)

    def guess(self, data, **kws):
        """Guess the parameters of the model."""
        pars = self.make_params()

        dim = data.dims[0]
        x_xr = data[dim].data

        # Estimate the Fermi level
        EF_estimate = data.estimate_EF()
        pars[f"{self.base_model_prefix}EF"].set(
            value=EF_estimate, max=max(x_xr), min=min(x_xr)
        )
        pars[f"{self.base_model_prefix}T"].set(
            value=data.attrs.get("temp_sample", 10), min=0, max=400, vary=False
        )

        # Guess background parameters
        cutoff = np.percentile(x_xr, 90)  # Take top 10% of data range
        bg_guess = data.sel({dim: slice(cutoff, None)}).quick_fit.linear()
        pars[f"{self.base_model_prefix}bg_slope"].set(value=bg_guess["slope"].data[()])
        pars[f"{self.base_model_prefix}bg_intercept"].set(
            value=bg_guess["intercept"].data[()]
        )

        # Guess DOS parameters
        cutoff = np.percentile(x_xr, 15)  # Take bottom 15% of data range
        dos_guess = data.sel({dim: slice(None, cutoff)}).quick_fit.linear()
        pars[f"{self.base_model_prefix}dos_slope"].set(value=dos_guess["slope"].data[()])
        pars[f"{self.base_model_prefix}dos_intercept"].set(
            value=dos_guess["intercept"].data[()]
        )

        # Set initial guess for Gaussian convolution
        pars["sigma_conv"].set(value=0.01, vary=True)

        return lm_models.update_param_vals(pars, self.base_model_prefix, **kws)


def _shirley_bg(data, num_avg=1, offset_start=0, offset_end=0, max_iterations=10):
    """Function to calculate the Shirley background of 1D data.

    Parameters
    ------------
    data : numpy.ndarray, list, xarray.DataArray
        The 1D data (y values) to find the Shirley background of.

    num_avg : int, optional
        The number of points to consider when calculating the average value of the data start and end points. Useful for
        noisy data. Defaults to 1.

    offset_start : float, optional
        The offset to subtract from the data start value. Useful when data range does not completely cover the start
        (left) tail of the peak. Defaults to 0.

    offset_end : float, optional
        The offset to subtract from the data end value. Useful when data range does not completely cover the end (right)
        tail of the peak. Defaults to 0.

    max_iterations : int, optional
        The maximum number of iterations to allow for convergence of Shirley background. Defaults to 10.

    Returns
    ------------
    Shirley_bkg : numpy.ndarray
        The Shirley background of the 1D data.

    Examples
    ------------
    Example usage is as follows::

        import peaks as pks

        S2p_XPS = pks.load('XPS1.ibw').DOS()

        # Extract Shirley background of the XPS scan
        S2p_XPS_Shirley_bkg = pks._Shirley(S2p_XPS)

        # Extract Shirley background of the XPS scan, using 3 points to calculate the average value of the data start
        and end points
        S2p_XPS_Shirley_bkg = pks._Shirley(S2p_XPS, num_avg=3)

        # Extract Shirley background of the XPS scan, applying an offset to the data start value
        S2p_XPS_Shirley_bkg = pks._Shirley(S2p_XPS, offset_start=0.3)

    """

    # Ensure data is of type numpy.ndarray
    if isinstance(data, np.ndarray):  # if data is a numpy array
        pass
    elif isinstance(data, xr.core.dataarray.DataArray):  # if data is a DataArray
        data = data.data
    elif isinstance(data, list):  # if data is a list
        data = np.array(data)
    else:
        raise Exception(
            "Inputted data must be a 1D numpy.ndarray, list or xarray.DataArray."
        )

    # Ensure data is 1D
    if len(data.shape) != 1:
        raise Exception(
            "Inputted data must be a 1D numpy.ndarray, list or xarray.DataArray."
        )

    # Ensure num_avg and max_iterations are integers
    try:
        num_avg = int(num_avg)
        max_iterations = int(max_iterations)
    except ValueError as e:
        raise Exception(
            "The inputs num_avg and max_iterations must both be integers"
        ) from e

    # Ensure offset_start and offset_end are floats
    try:
        offset_start = float(offset_start)
        offset_end = float(offset_end)
    except ValueError as e:
        raise Exception(
            "The inputs offset_start and offset_end must both be floats"
        ) from e

    # Get number of points in data and define tolerance
    num_points = len(data)
    tolerance = 1e-5

    # Determine start and end limits of Shirley background
    y_start = data[0:num_avg].mean() - offset_start
    y_end = data[(num_points - num_avg) : num_points].mean() - offset_end

    # Initialise the bkg shape B, where total Shirley bkg is given by Shirley_bkg = y_end + B
    B = np.zeros(data.shape)

    # First B value is equal to y_start - y_end, i.e. Shirley_bkg[0] = y_start as expected
    B[0] = y_start - y_end

    # Define function to determine Shirley bkg
    def dk(i):
        return sum(0.5 * (data[i:-1] + data[i + 1 :] - 2 * y_end - B[i:-1] + B[i + 1 :]))

    # Perform iterative procedure to converge to Shirley bkg, stopping if maximum number of iterations is reached
    num_iterations = 0
    while num_iterations < max_iterations:
        # Calculate new k = (y_start - y_end) / (int_(xl)^(xr) J(x') - y_end - B(x') dx')
        k_sum = sum(0.5 * (data[:-1] + data[1:] - 2 * y_end - B[:-1] + B[1:]))
        k = (y_start - y_end) / k_sum
        # Calculate new B
        y_sum = np.array(list(map(dk, range(0, num_points))))
        new_B = k * y_sum
        # If new_B is close to B (within tolerance), stop the loop
        if sum(abs(new_B - B)) < tolerance:
            B = new_B
            break
        else:
            B = new_B
        num_iterations += 1

    # Raise an error if the maximum allowed number of iterations is exceeded
    if num_iterations >= max_iterations:
        raise Exception(
            "Maximum number of iterations exceeded before convergence of Shirley background was achieved."
        )

    # Determine Shirley bkg
    Shirley_bkg = y_end + B

    return Shirley_bkg



import re
from datetime import datetime

import numpy as np
import pint_xarray

from peaks.core.fileIO.base_arpes_data_classes.base_arpes_data_class import (
    BaseARPESDataLoader,
)
from peaks.core.fileIO.loc_registry import register_loader
from peaks.core.utils.misc import analysis_warning

ureg = pint_xarray.unit_registry


@register_loader
class MBSDataLoader(BaseARPESDataLoader):
    """Generic data loader for MBS data.

    Notes
    ------------
    This class is intended to be subclassed to provide specific loaders for different locs using MBS data formats.

    Subclasses should define the `_MBS_metadata_key_mappings` class variable to map any custom metadata fields or
    fixed values to the standard peaks metadata keys. See the docstrings for `_MBS_metadata_dict_keys_to_peaks_keys`
    for more information.

    Subclasses should also define the `_MBS_metadata_units` dictionary to map any fixed units that are not determined
    as part of the metadata loader.

    """

    _loc_name = "MBS"
    _loc_description = (
        "Loader for data acquired using the MBS A1 soft acquisition software."
    )
    _loc_url = "https://www.mbscientific.se/"
    # Dictionary to overwrite or add to default metadata keys:
    _MBS_metadata_key_mappings = {}  # mappings from MBS metadata key to peaks key
    _MBS_metadata_units = {}  # standard MBS units
    _analyser_name_conventions = {
        "deflector_parallel": "DeflY",
        "deflector_perp": "DeflX",
    }

    @classmethod
    def _load_data(cls, fpath, lazy):
        """Load MBS data from different file types."""
        handlers = {
            "txt": cls._load_from_txt,
            "krx": cls._load_from_krx,
        }
        ext = fpath.split(".")[-1]
        if ext not in handlers:
            raise ValueError(
                f"File extension {ext} is not supported for loader {cls.__name__}. Should be one of "
                f"{list(handlers.keys())}"
            )

        # Get the metadata here returning with keys in MBS format - this as needed for parsing the core data.
        metadata_dict_MBS_keys = cls._load_metadata(fpath, return_in_MBS_format=True)
        # Cahce it in the metadata cache to avoid having to load it again later
        cls._metadata_cache[fpath] = metadata_dict_MBS_keys

        # Load data
        return handlers[ext](fpath, metadata_dict_MBS_keys)

    @classmethod
    def _load_from_txt(cls, fpath, metadata_dict_MBS_keys):
        """Load data from a .txt file."""
        # Give a legacy warning
        analysis_warning(
            "Loading MBS data from .txt file. This is a legacy format and only basic loading is supported. "
            "It is strongly encouraged to save data using the .krx format from the MBS software.",
            "warning",
            "Legacy format - MBS .txt file loader",
        )

        # Load file data
        file_data = np.loadtxt(
            fpath, skiprows=int(metadata_dict_MBS_keys["metadata_lines_length"])
        )
        spectrum = file_data[:, 1:]
        eV_values = file_data[:, 0]
        eV_units = "eV"
        y_scale_values = np.arange(
            float(metadata_dict_MBS_keys["ScaleMin"]),
            float(metadata_dict_MBS_keys["ScaleMax"]),
            float(metadata_dict_MBS_keys["ScaleMult"]),
        )
        y_scale_name, y_scale_units = cls._parse_axis_name_and_units(
            metadata_dict_MBS_keys["ScaleName"]
        )
        y_scale_name = (
            "theta_par" if y_scale_name.lower() in ["angle", "y angle"] else "y_scale"
        )

        return {
            "spectrum": spectrum,
            "dims": ["eV", y_scale_name],
            "coords": {"eV": eV_values, y_scale_name: y_scale_values},
            "units": {
                "eV": eV_units,
                y_scale_name: y_scale_units,
                "spectrum": "counts",
            },
        }

    @classmethod
    def _load_from_krx(cls, fpath, metadata_dict_MBS_keys):
        """Load data from krx format."""

        with open(fpath, "rb") as f:
            # Determine whether the file is 32-bit or 64-bit. The data type is little endian, so read initially as
            # 32 bit, but if either of the first 2 32-bit words are 0, then the file is 64-bit.
            dtype_identifier = np.fromfile(f, dtype="<i4", count=2)
            if 0 in dtype_identifier:  # File is 64 bit
                dtype = "<i8"  # 8-byte signed integer (little endian)
            else:  # File is 32 bit
                dtype = "<i4"  # 4-byte signed integer (little endian)

            # Set reading to the start of the file
            f.seek(0)

            # Read the pointer array size, which is the first word of the array
            pointer_array_size = np.fromfile(f, dtype=dtype, count=1)

            # Determine the number of images in the file
            num_images = int(pointer_array_size[0] / 3)

            # Sequentially read the image positions and sizes
            image_pos = []
            Y_size = []
            X_size = []
            for _i in range(num_images):
                # Extract position in array of the images in 32-bit integers
                image_pos.append(np.fromfile(f, dtype=dtype, count=1)[0])
                # Extract Y size of array (angular direction)
                Y_size.append(np.fromfile(f, dtype=dtype, count=1)[0])
                # Extract X size of array (energy direction)
                X_size.append(np.fromfile(f, dtype=dtype, count=1)[0])

            # Read more file information to identify the scan type
            scan_identifier = np.fromfile(f, dtype=dtype, count=1)[
                0
            ]  # 5 for spin, 4 for ARPES
            # Calculate the array size
            array_size = X_size[0] * Y_size[0]
            # Set file position to the first header
            f.seek((image_pos[0] + array_size + 1) * 4)

            # Extract the kinetic energy values
            KE_values = np.linspace(
                float(metadata_dict_MBS_keys["Start K.E."]),
                float(metadata_dict_MBS_keys["End K.E."]),
                X_size[0],
            )

            # Extract the theta_par values
            if scan_identifier != 5:  # ARPES scans, extract analyser MCP angular scale
                y_scale_values = np.linspace(
                    float(metadata_dict_MBS_keys["ScaleMin"]),
                    float(metadata_dict_MBS_keys["ScaleMax"]),
                    Y_size[0],
                )
            else:  # Spin ARPES scans, extract spin MCP angular scale
                y_scale_values = np.linspace(
                    float(metadata_dict_MBS_keys["S0ScaleMin"]),
                    float(metadata_dict_MBS_keys["S0ScaleMax"]),
                    Y_size[0],
                )
            y_scale_name, y_scale_units = cls._parse_axis_name_and_units(
                metadata_dict_MBS_keys["ScaleName"]
            )
            y_scale_name = (
                "theta_par"
                if y_scale_name.lower() in ["angle", "y angle"]
                else "y_scale"
            )

            # If there is a single image, load 2D spectrum
            if num_images == 1:
                # Set read position in the file to the image location
                f.seek(image_pos[0] * 4)
                # Read the image spectrum (images written as 32-bit words even in 64-bit format .krx file)
                spectrum = np.fromfile(f, dtype="<i4", count=array_size)
                # Reshape spectrum into the desired data structure
                spectrum = np.reshape(spectrum, [Y_size[0], X_size[0]])

                return {
                    "spectrum": spectrum,
                    "dims": [
                        y_scale_name,
                        "eV",
                    ],
                    "coords": {"eV": KE_values, y_scale_name: y_scale_values},
                    "units": {
                        "eV": "eV",
                        y_scale_name: y_scale_units,
                        "spectrum": "counts",
                    },
                }

            # If there are multiple images, load the spectrum as a data cube
            else:
                # Define the spectrum in the order [mapping_dim, theta_par, eV]
                spectrum = np.zeros((num_images, Y_size[0], X_size[0]))

                # Loop through the images and fill in spectrum
                for i, pos in enumerate(image_pos):
                    # Set the read position in the file to the current image location
                    f.seek(pos * 4)

                    # Read the current image data (images written as 32-bit words even in 64-bit format .krx file)
                    current_image_data = np.fromfile(f, dtype="<i4", count=array_size)

                    # Reshape the current image data into the desired data structure, and fill entries in spectrum
                    spectrum[i, :, :] = np.reshape(
                        current_image_data, [Y_size[0], X_size[0]]
                    )

                # If scan type is an ARPES map of some form, extract the deflector angular values
                if scan_identifier == 4:  # ARPES Deflector scan
                    mapping_values = np.linspace(
                        float(metadata_dict_MBS_keys["MapStartX"]),
                        float(metadata_dict_MBS_keys["MapEndX"]),
                        num_images,
                    )
                    mapping_label = "deflector_perp"
                # If scan type is a spin scan, extract the spin rotation angle values
                elif scan_identifier == 5:  # Spin scan
                    mapping_values = []
                    for i in range(num_images):
                        current_spin_rot_angle = float(
                            metadata_dict_MBS_keys["SpinComp#" + str(i)]
                            .split(",", 1)[1]
                            .split(">", 1)[0]
                        )
                        mapping_values.append(current_spin_rot_angle)
                    mapping_label = "spin_rot_angle"

                return {
                    "spectrum": spectrum,
                    "dims": [
                        mapping_label,
                        y_scale_name,
                        "eV",
                    ],
                    "coords": {
                        "eV": KE_values,
                        y_scale_name: y_scale_values,
                        mapping_label: mapping_values,
                    },
                    "units": {
                        "eV": "eV",
                        y_scale_name: y_scale_units,
                        mapping_label: "deg",
                        "spectrum": "counts",
                    },
                }

    @classmethod
    def _load_metadata(cls, fpath, return_in_MBS_format=False):
        """Load metadata from an MBS file."""
        # Check if there is a cached version (only cache in this loader with keys in MBS format)
        metadata_dict_MBS_keys = cls._metadata_cache.get(fpath)

        # If no metadata in the cache, load it
        if not metadata_dict_MBS_keys:
            handlers = {
                "txt": cls._load_MBS_metadata_txt,
                "krx": cls._load_MBS_metadata_krx,
            }
            ext = fpath.split(".")[-1]
            metadata_lines = handlers[ext](fpath)
            metadata_dict_MBS_keys = cls._MBS_metadata_to_dict_w_MBS_keys(metadata_lines)

        if return_in_MBS_format:
            return metadata_dict_MBS_keys

        # Convert the metadata to the peaks convention and return
        return cls._MBS_metadata_dict_keys_to_peaks_keys(metadata_dict_MBS_keys)

    @staticmethod
    def _load_MBS_metadata_txt(fpath):
        """Extract the lines containing metadata in an MBS format .txt file.

        Returns
        ------------
        metadata_lines : list
            Lines extracted from the file containing the metadata.

        """
        # Open the file and extract the lines containing metadata
        with open(fpath) as f:
            metadata_lines = []
            while True:
                line = f.readline()
                metadata_lines.append(line.rstrip("\n"))
                # When the line starting with '[Data' is encountered, metadata has ended and scan data has begun
                if "DATA:" in line:
                    break
            # Manually add the lengths of the metadata section as needed for .txt loader
            metadata_lines.append(f"metadata_lines_length\t{len(metadata_lines)}")
        return metadata_lines

    @staticmethod
    def _load_MBS_metadata_krx(fpath):
        """Extract the lines containing metadata in an MBS format .krx file.

        Returns
        ------------
        metadata_lines : list
            Lines extracted from the file containing the metadata.

        """
        # Open the file in read mode and extract metalines
        with open(fpath, "rb") as f:
            # Determine whether the file is 32-bit or 64-bit. The data type is little endian, so read initially as
            # 32 bit, but if either of the first 2 32-bit words are 0, then the file is 64-bit.
            dtype_identifier = np.fromfile(f, dtype="<i4", count=2)
            if 0 in dtype_identifier:  # File is 64 bit
                dtype = "<i8"  # 8-byte signed integer (little endian)
            else:  # File is 32 bit
                dtype = "<i4"  # 4-byte signed integer (little endian)

            # Set reading to the start of the file
            f.seek(0)

            # Read the pointer array size, which is the first word of the array. Pointer array size is not used here,
            # but is read so that the next variable read is image position
            pointer_array_size = np.fromfile(f, dtype=dtype, count=1)  # noqa: F841

            # Extract first image position and array size to know where in the file to find metadata
            image_pos = np.fromfile(f, dtype=dtype, count=1)[0]
            Y_size = np.fromfile(f, dtype=dtype, count=1)[0]
            X_size = np.fromfile(f, dtype=dtype, count=1)[0]
            array_size = X_size * Y_size

            # Set file position to the first header
            f.seek((image_pos + array_size + 1) * 4)

            # Read the header (allowing up to 1800 bytes) containing metadata and convert into ascii format
            header = f.read(1800).decode("ascii")

            # Shorten header to required length (i.e. up to where scan data starts)
            header = header.split("\r\nDATA:")[0]

            return header.split("\r\n")

    @staticmethod
    def _parse_axis_name_and_units(string):
        match = re.match(r"([^\(]+)\(([^)]+)\)", string)
        if match:
            name, units = match.groups()
            return name.strip(), units.strip().lower()
        else:
            return string.strip(), None

    @staticmethod
    def _MBS_metadata_to_dict_w_MBS_keys(metadata_lines):
        """Convert metadata lines to a dictionary of key-value pairs with the keys being the metadata entries
        as they appear in the MBS metadata (i.e. not yet in :class:`peaks` convention).

        Returns
        ------------
        dict
            Dictionary of metadata key-value pairs with keys in MBS format.
        """

        meta_dict = {
            line.split("\t", 1)[0]: line.split("\t", 1)[1]
            for line in metadata_lines
            if "\t" in line
        }
        for k, v in meta_dict.items():
            try:
                meta_dict[k] = float(v)
            except ValueError:
                pass

        return meta_dict

    @classmethod
    def _MBS_metadata_dict_keys_to_peaks_keys(cls, metadata_dict_MBS_keys):
        """Extract metadata values in peaks conventions and assign units where appropriate.

        Parameters
        ------------
        metadata_dict_MBS_keys : dict
            Dictionary of metadata key-value pairs with keys in MBS format.

        Returns
        ------------
        metadata_dict : dict
            Dictionary of metadata key-value pairs with keys in peaks format.

        Notes
        ------------
        The extraction process is based on the mappings defined in the dictionaries `standard_keys` and
        `standard_units`. The entries of these dictionaries are updated from the class variables
        `_MBS_metadata_key_mappings` and `_MBS_metadata_units` respectively, and so subclasMBS can overwrite and
        extend these defaults by specifying the appropriate mappings in these class variables.

        The extraction process supports different types of keys:
            - If a single key is given, the value is extracted directly.
            - If a list of keys is given, the function tries to extract values for all keys and returns a list or
            array of all non-None values.
            - If a callable is given, the function calls it, passing it the metadata dictionary and returns the result.
            - A constant value can be given by defining a simple lambda function, for example `lambda x: -1`.

        """

        def _parse_KE(metadata_dict):
            ana_mode = metadata_dict.get("AcqMode")
            if ana_mode == "Fixed":
                return metadata_dict.get("Center K.E.")
            return np.asarray(
                [metadata_dict.get("Start K.E."), metadata_dict.get("End K.E.")]
            )

        # Define the mapping assuming the standard keys for MBS metadata
        standard_keys = {
            "analyser_model": lambda x: "A1",
            "analyser_slit_width": None,
            "analyser_slit_width_identifier": None,
            "analyser_eV": _parse_KE,
            "analyser_eV_type": lambda x: "kinetic",
            "analyser_step_size": "Step Size",
            "analyser_PE": lambda x: float(x.get("Pass Energy")[2:]),
            "analyser_sweeps": "No Scans",
            "analyser_dwell": "Frames Per Step",
            "analyser_lens_mode": "Lens Mode",
            "analyser_acquisition_mode": "AcqMode",
            "analyser_polar": None,
            "analyser_tilt": None,
            "analyser_azi": None,
            "analyser_deflector_parallel": "DeflY",
            "analyser_deflector_perp": "DeflX",
            "timestamp": lambda x: datetime.strptime(
                x.get("TIMESTAMP:"), "%d/%m/%Y %H:%M"
            ).strftime("%Y-%m-%d %H:%M:%S"),
        }
        # Define standard units
        standard_units = {
            "analyser_model": None,
            "analyser_slit_width": None,
            "analyser_slit_width_identifier": None,
            "analyser_eV": "eV",
            "analyser_step_size": "eV",
            "analyser_PE": "eV",
            "analyser_sweeps": None,
            "analyser_dwell": "ms",
            "analyser_lens_mode": None,
            "analyser_acquisition_mode": None,
            "analyser_polar": "deg",
            "analyser_tilt": "deg",
            "analyser_azi": "deg",
            "analyser_deflector_parallel": "deg",
            "analyser_deflector_perp": "deg",
            "photon_hv": "eV",
        }

        # If custom key mappings or units have been supplied, update the standard keys
        standard_keys.update(cls._MBS_metadata_key_mappings)
        standard_units.update(cls._MBS_metadata_units)

        # Extract metadata values and give them units where appropriate
        # - if a single key is given, extract the value directly
        # - if a list of keys is given, try and extract for all and return all non-None values
        # - if a callable is given, call the function with the metadata dictionary and return the result
        # - add units if given and extracted value is not None
        metadata_dict = {}
        units_failure = []

        def try_parse_to_float(value):
            try:
                return float(value)
            except (ValueError, TypeError):
                return value

        # Handle extracting the metadata based on the arguments in the standard_keys dictionary
        for peaks_key, MBS_key in standard_keys.items():
            value = None
            if isinstance(MBS_key, str):
                value = try_parse_to_float(metadata_dict_MBS_keys.get(MBS_key))
            elif callable(MBS_key):
                value = MBS_key(try_parse_to_float(metadata_dict_MBS_keys))
            elif isinstance(MBS_key, list):
                value = [
                    try_parse_to_float(metadata_dict_MBS_keys.get(key))
                    for key in MBS_key
                    if metadata_dict_MBS_keys.get(key) is not None
                ]

                if len(value) == 1:
                    value = value[0]
                elif len(value) == 0:
                    value = None
                else:
                    try:
                        value = np.asarray(value)
                    except (ValueError, TypeError):
                        value = str(value)

            # Try to add units if they are given and the value is not None
            if value is not None and standard_units.get(peaks_key):
                try:
                    value = value * ureg(standard_units[peaks_key])
                except TypeError:
                    units_failure.append(peaks_key)  # Keep track of errors
            metadata_dict[peaks_key] = value

        # Warn if any units failed to be added
        if units_failure:
            analysis_warning(
                f"Failed to add units to the following metadata items: {units_failure}",
                "warning",
                "Failed to add units to metadata",
            )

        return metadata_dict



"""Functions for the 4D interactive display panel."""

import sys
from functools import partial

import numpy as np
import pyperclip
import pyqtgraph as pg
from PyQt6 import QtCore, QtWidgets
from PyQt6.QtWidgets import (
    QApplication,
    QHBoxLayout,
    QLabel,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

from ..GUI_utils import (
    Crosshair,
    KeyPressGraphicsLayoutWidget,
)
from .disp_2d import _Disp2D


def _disp_4d(data, primary_dim):
    """Display a 4D interactive display panel.

    Parameters
    ------------
    data : list or xarray.DataArray
         Either a single 2D :class:`xarray.DataArray` or a list of 2D :class:`xarray.DataArray` objects.

    primary_dim : tuple of str or list of str
        The primary dimensions for the viewer, will be shown on main data explorer map panel.

    """
    global app  # Ensure the QApplication instance is not garbage collected
    app = QApplication.instance()
    if app is None:
        app = QApplication(sys.argv)

    viewer = _Disp4D(data, primary_dim)
    viewer.show()
    app.exec()


class _Disp4D(QtWidgets.QMainWindow):
    def __init__(self, data, primary_dim):
        super().__init__()

        # Parse data
        if primary_dim:  # Set correct primary axis
            data = data.transpose(primary_dim[0], primary_dim[1], ...)
        self.data = data

        # Set some options
        self.xh_brush = (255, 0, 255, 50)

        # Set up the GUI
        self._init_UI()  # Initialize the layout

    # ##############################
    # GUI layout
    # ##############################
    def _init_UI(self):
        self.setWindowTitle("Display Panel")
        self.setGeometry(100, 100, 1200, 800)

        # Main window layout
        central_widget = QTabWidget()
        self.setCentralWidget(central_widget)
        explorer_tab = QWidget()
        central_widget.addTab(explorer_tab, "Data Explorer")
        ROI_tab = QWidget()
        central_widget.addTab(ROI_tab, "Region of Interest")

        self._init_data()  # Read some basic parameters from the dataset
        self._build_explorer_tab(explorer_tab)
        self._build_roi_tab(ROI_tab)
        self._connect_signals()  # Connect signals
        # self._build_plot_layout(layout)  # Build the basic plot layout
        # self._build_controls_layout(layout)  # Build the control panel layout
        # self._build_menu()  # Add the menu

    def _build_explorer_tab(self, explorer_tab):
        """Build the data explorer tab layout."""
        layout = QHBoxLayout()
        explorer_tab.setLayout(layout)
        self._build_primary_dims_explorer_column(layout)
        # Add a disp2d panel for the dim2-dim3 data
        self.data_explorer_disp2d = self._build_data_explorer_2D_column(
            layout, self.data.isel({self.dims[0]: 0, self.dims[1]: 0})
        )

    def _build_roi_tab(self, roi_tab):
        """Build the region of interest tab layout."""
        layout = QHBoxLayout()
        roi_tab.setLayout(layout)
        self._build_primary_dims_roi_column(layout)
        # Add a disp2d panel for the dim2-dim3 data
        self.roi_disp2d = self._build_data_explorer_2D_column(
            layout, self.data_secondary_tot
        )
        # Add a ROI to the disp2d panel
        self._add_ROI_to_disp2d()

    def _build_primary_dims_explorer_column(self, layout):
        """Build left column - primary dims explorer"""
        left_column_container = QWidget()
        layout.addWidget(left_column_container)
        left_column_container.setMaximumWidth(500)
        left_column_layout = QVBoxLayout(left_column_container)

        _, _, self.primary_dims_xh, _, self.primary_dims_explorer_layout = (
            self._build_primary_dims_explorer_plot(left_column_layout)
        )
        self._build_primary_dims_explorer_controls(left_column_layout)

        left_column_layout.addStretch()

    def _build_primary_dims_roi_column(self, layout):
        """Build left column - primary dims ROI"""
        left_column_container = QWidget()
        layout.addWidget(left_column_container)
        left_column_container.setMaximumWidth(500)
        left_column_layout = QVBoxLayout(left_column_container)

        (
            self.primary_dims_roi_plot,
            self.primary_dims_roi_plot_image_item,
            xh,
            self.primary_dims_roi_plot_colorbar,
            _,
        ) = self._build_primary_dims_explorer_plot(left_column_layout)
        xh.set_visible(False)
        # Add a ROI
        self.roi_dim01 = pg.PolyLineROI(
            positions=self._set_ROI_default_pos(0, 1),
            pen=pg.mkPen((204, 51, 153), width=2.5),
            handlePen=pg.mkPen((204, 51, 153), width=2),
            closed=True,
        )
        self.primary_dims_roi_plot.addItem(self.roi_dim01)
        self._build_roi_controls(left_column_layout)

        left_column_layout.addStretch()

    def _build_primary_dims_explorer_plot(self, left_column_layout):
        # Create a GraphicsLayoutWidget
        primary_dims_explorer_layout = (
            KeyPressGraphicsLayoutWidget._KeyPressGraphicsLayoutWidget()
        )
        primary_dims_explorer_layout.setMaximumWidth(500)
        primary_dims_explorer_layout.viewport().setAttribute(
            QtCore.Qt.WidgetAttribute.WA_AcceptTouchEvents, False
        )
        left_column_layout.addWidget(primary_dims_explorer_layout)
        primary_dims_explorer_plot = primary_dims_explorer_layout.addPlot(1, 0, 1, 1)
        # Add primary dims image total data to plot
        primary_dims_explorer_plot_image_item = pg.ImageItem(
            self.data_primary_tot.values, axisOrder="row-major"
        )
        primary_dims_explorer_plot.addItem(primary_dims_explorer_plot_image_item)

        # Transform to match data coordinates
        primary_dims_explorer_plot_image_item.setTransform(
            pg.QtGui.QTransform(
                self.step_sizes[1],
                0,
                0,
                0,
                self.step_sizes[0],
                0,
                self.ranges[1][0],
                self.ranges[0][0],
                1,
            )
        )

        # Set view limits
        xmin, xmax = self.ranges[1]
        ymin, ymax = self.ranges[0]
        primary_dims_explorer_plot.getViewBox().setLimits(
            xMin=xmin, xMax=xmax, yMin=ymin, yMax=ymax
        )

        # Set axis labels
        dim_labels = []
        for i in range(2):
            dim_units = self.data.coords[self.dims[i]].attrs.get("units")
            dim_labels.append(
                f"{self.dims[i]} ({dim_units})" if dim_units else self.dims[i]
            )

        primary_dims_explorer_plot.setLabel("bottom", dim_labels[1])
        primary_dims_explorer_plot.setLabel("left", dim_labels[0])

        # Add a crosshair
        primary_dims_xh = Crosshair._Crosshair(
            primary_dims_explorer_plot,
            pos=(
                self.coords[0][0],
                self.coords[1][0],
            ),
            dim0_width=self.step_sizes[0],
            dim1_width=self.step_sizes[1],
            brush=self.xh_brush,
            bounds=[
                self.ranges[0],
                self.ranges[1],
            ],
            axisOrder="row-major",
        )

        # Add colour bar
        crange = (
            float(self.data_primary_tot.min()),
            float(self.data_primary_tot.max()),
        )
        cmap = pg.colormap.get("cividis", source="matplotlib")
        colorbar = pg.ColorBarItem(
            label=self.data.name,
            colorMap=cmap,
            colorMapMenu=True,
            limits=crange,
            rounding=min(abs(crange[1] - crange[0]) / 2000, 1),
            interactive=True,
            orientation="h",
            values=crange,
        )
        colorbar.setImageItem(
            primary_dims_explorer_plot_image_item,
            insert_in=primary_dims_explorer_plot,
        )
        primary_dims_explorer_plot_image_item.setColorMap(cmap)
        primary_dims_explorer_layout.addItem(colorbar, row=0, col=0)

        return (
            primary_dims_explorer_plot,
            primary_dims_explorer_plot_image_item,
            primary_dims_xh,
            colorbar,
            primary_dims_explorer_layout,
        )

    def _build_primary_dims_explorer_controls(self, left_column_layout):
        """Add controls to the primary dims explorer plot."""
        controls_layout = QVBoxLayout()
        left_column_layout.addLayout(controls_layout)
        self.dim01_sliders = []
        self.dim01_binning_boxes = []
        for i in range(2):
            hbox = QHBoxLayout()
            controls_layout.addLayout(hbox)
            label = QLabel(self.dims[i])
            hbox.addWidget(label)
            # Selected slice control
            slider = QtWidgets.QSlider(QtCore.Qt.Orientation.Horizontal)
            slider.setTickPosition(QtWidgets.QSlider.TickPosition.TicksBelow)
            slider.setRange(0, len(self.coords[i]) - 1)
            slider.setValue(0)
            hbox.addWidget(slider)
            self.dim01_sliders.append(slider)
            # Binning control
            label = QLabel("Bin:")
            hbox.addWidget(label)
            binning_box = QtWidgets.QDoubleSpinBox()
            binning_box.setRange(1, 50)
            binning_box.setSingleStep(1)
            binning_box.setValue(1)
            binning_box.setDecimals(0)
            hbox.addWidget(binning_box)
            self.dim01_binning_boxes.append(binning_box)

        # Add a text box to display the current slice
        self.primary_dims_cursor_stats = QLabel()
        self.primary_dims_cursor_stats.setStyleSheet(
            "QLabel { background-color : black; color : white; }"
        )
        self.primary_dims_cursor_stats.setAlignment(
            QtCore.Qt.AlignmentFlag.AlignVCenter | QtCore.Qt.AlignmentFlag.AlignLeft
        )
        self.primary_dims_cursor_stats.setWordWrap(True)
        self.primary_dims_cursor_stats.setMinimumHeight(60)
        controls_layout.addWidget(self.primary_dims_cursor_stats)
        self._update_primary_dim_cursor_stats()

    def _build_roi_controls(self, left_column_layout):
        """Add controls to the ROI tab."""
        controls_layout = QHBoxLayout()
        left_column_layout.addLayout(controls_layout)
        vbox = QVBoxLayout()
        controls_layout.addLayout(vbox)
        self.roi_list_box = QtWidgets.QListWidget()
        for key in self.ROI_store.keys():
            self.roi_list_box.addItem(key)
        self.roi_list_box.setSelectionMode(
            QtWidgets.QAbstractItemView.SelectionMode.MultiSelection
        )
        vbox.addWidget(self.roi_list_box)
        self.clear_ROI_store_button = QtWidgets.QPushButton("Clear all ROIs")
        vbox.addWidget(self.clear_ROI_store_button)

        # Main buttons
        button_layout = QVBoxLayout()
        controls_layout.addLayout(button_layout)
        # Add buttons
        self.setROI_button = QtWidgets.QPushButton("Set ROI from plot")
        button_layout.addWidget(self.setROI_button)
        hbox = QHBoxLayout()
        button_layout.addLayout(hbox)
        hbox.addWidget(QLabel("|---->"))
        self.select_from_dim01_selector = QtWidgets.QRadioButton(
            f"From [{self.dims[0]}, {self.dims[1]}]"
        )
        self.select_from_dim01_selector.setChecked(True)
        hbox.addWidget(self.select_from_dim01_selector)
        hbox.addStretch()
        hbox = QHBoxLayout()
        button_layout.addLayout(hbox)
        hbox.addWidget(QLabel("<----|"))
        self.select_from_dim23_selector = QtWidgets.QRadioButton(
            f"From [{self.dims[2]}, {self.dims[3]}]"
        )
        hbox.addWidget(self.select_from_dim23_selector)
        hbox.addStretch()
        self.ROI_name_box = QtWidgets.QLineEdit()
        button_layout.addWidget(self.ROI_name_box)
        self.addROI_button = QtWidgets.QPushButton("Add to list")
        button_layout.addWidget(self.addROI_button)
        self.removeROI_button = QtWidgets.QPushButton("Remove from list")
        button_layout.addWidget(self.removeROI_button)
        self.setROI_from_list_button = QtWidgets.QPushButton("Set ROI from list")
        button_layout.addWidget(self.setROI_from_list_button)
        self.copy_ROI_store_button = QtWidgets.QPushButton("Copy ROI store")
        button_layout.addWidget(self.copy_ROI_store_button)
        self.resetplots_button = QtWidgets.QPushButton("Reset Plots")
        button_layout.addWidget(self.resetplots_button)
        button_layout.addStretch()

    def _build_data_explorer_2D_column(self, layout, data_to_plot):
        """Right column - data explorer utilising the disp2d panel"""
        right_column_container = QWidget()
        layout.addWidget(right_column_container)
        right_column = QVBoxLayout()
        layout.addLayout(right_column)
        # Add the 2D panel - initialise with the first slice
        data_explorer_disp2d = _Disp2D([data_to_plot], None, None)
        data_explorer_widget = data_explorer_disp2d.centralWidget()
        right_column.addWidget(data_explorer_widget)
        # Hide the unnecessary controls from disp2d
        data_explorer_disp2d.align_button.setVisible(False)
        data_explorer_disp2d.copy_button.setVisible(False)
        data_explorer_disp2d.scans_list.setVisible(False)

        return data_explorer_disp2d

    # ##############################
    # Signals
    # ##############################
    def _connect_signals(self):
        """Connect signals."""
        # Connect signals for primary dims explorer

        # Set sliders position based on xh
        signal = self.primary_dims_xh.xh.sigPositionChanged
        signal.connect(self._update_dim01_slider_position)

        for i in range(2):
            # Set binning for cursors
            signal = self.dim01_binning_boxes[i].valueChanged
            signal.connect(partial(self._update_dim01_binning, i))
            # Update plot when slider changes
            signal = self.dim01_sliders[i].valueChanged
            signal.connect(self._update_data_explorer_disp2d)
            # Move xh when slider dragged
            signal = self.dim01_sliders[i].sliderMoved
            signal.connect(self._update_primary_dim_xhs_from_sliders)

        # Set key press signals
        self.primary_dims_explorer_layout.keyPressed.connect(self._key_press_event)

        # Connect signals for ROI tab
        self.resetplots_button.clicked.connect(self._reset_roi_plots)
        self.setROI_button.clicked.connect(self._set_ROI_from_plot)
        self.addROI_button.clicked.connect(self._add_ROI_to_list)
        self.removeROI_button.clicked.connect(self._remove_selected_ROI_from_list)
        self.clear_ROI_store_button.clicked.connect(self._remove_all_ROIs_from_list)
        self.setROI_from_list_button.clicked.connect(self._set_ROI_from_list)
        self.roi_list_box.itemSelectionChanged.connect(self._select_ROI_from_list)
        self.copy_ROI_store_button.clicked.connect(
            lambda: pyperclip.copy(f".attrs['ROI']={str(self.ROI_store)}")
        )

    # ##############################
    # Data handling
    # ##############################
    def _init_data(self):
        """Extract some core parameters from the data."""
        self.dims = self.data.dims
        self.coords = [self.data.coords[dim].values for dim in self.dims]
        self.coarsened_coords = [coord for coord in self.coords[:2]]
        self.step_sizes = [
            (coord[-1] - coord[0]) / (len(coord) - 1) for coord in self.coords
        ]
        self.ranges = [
            (min(coords) - self.step_sizes[i] / 2, max(coords) + self.step_sizes[i] / 2)
            for i, coords in enumerate(self.coords)
        ]
        self.data_span = [abs(range[1] - range[0]) for range in self.ranges]
        self.data_primary_tot = self.data.mean([self.dims[2], self.dims[3]]).compute()
        self.data_secondary_tot = self.data.mean([self.dims[0], self.dims[1]]).compute()

        # Check the real axis name if possible
        self.dim_labels = []
        for dim in self.dims:
            real_axis_name = None
            if hasattr(self.data.metadata, "manipulator"):
                if hasattr(self.data.metadata.manipulator, dim):
                    real_axis_name = getattr(
                        self.data.metadata.manipulator, dim
                    ).local_name
            self.dim_labels.append(
                f"{dim} [{real_axis_name}]" if real_axis_name else dim
            )

        # Make a ROI store
        self.ROI_store = self.data.attrs.get("ROI", {})
        self.display_ROIs_01 = []
        self.display_ROIs_23 = []

    def _update_primary_dim_cursor_stats(self):
        """Update the cursor stats display."""

        r, g, b = self.xh_brush[:3]
        cursor_text = f"<span style='color:#{r:02x}{g:02x}{b:02x}; font-size:15px'>"
        for i, pos in enumerate(self.primary_dims_xh.get_pos()):
            cursor_text += f"&nbsp;&nbsp;{self.dim_labels[i]}:&nbsp;&nbsp; {pos:.2f}<br>"
        cursor_text += "</span>"
        self.primary_dims_cursor_stats.setText(cursor_text)

    def _update_dim01_slider_position(self):
        """Update the slider positions based on the crosshair position."""
        pos = self.primary_dims_xh.get_pos()
        for i in range(2):
            index = self._find_nearest_sorted(self.coarsened_coords[i], pos[i])
            self.dim01_sliders[i].setValue(index)
        self._update_primary_dim_cursor_stats()

    def _update_dim01_binning(self, dim_no):
        """Update the dim0 and dim1 binning."""
        binning = int(self.dim01_binning_boxes[dim_no].value())
        dim = self.dims[dim_no]
        # Update the xh
        getattr(self.primary_dims_xh, f"set_dim{dim_no}_width")(
            self.step_sizes[dim_no] * binning
        )
        # Make a new rolling coordinate
        self.coarsened_coords[dim_no] = (
            self.data[self.dims[dim_no]]
            .rolling(dim={dim: binning}, center=True)
            .mean()
            .dropna(dim)
            .values
        )
        self.dim01_sliders[dim_no].setRange(0, len(self.coarsened_coords[dim_no]) - 1)
        self._update_data_explorer_disp2d()

    def _update_data_explorer_disp2d(self):
        """Update the 2D data explorer plot."""
        # st = time.perf_counter()
        # Find the indices to slice over
        pos = self.primary_dims_xh.get_pos()
        indices = [
            self._find_nearest_n_indexes(
                self.coords[i], pos[i], int(self.dim01_binning_boxes[i].value())
            )
            for i in range(2)
        ]
        # Slice (and if needed average) the data
        data_2d = self.data.data[indices[0], indices[1], :, :]
        if data_2d.ndim == 4:
            data_2d = data_2d.mean((0, 1))
        elif data_2d.ndim == 3:
            data_2d = data_2d.mean(0)

        if not isinstance(data_2d, np.ndarray):
            data_2d = data_2d.compute()

        self.data_explorer_disp2d.data_arrays[0].data = data_2d
        self.data_explorer_disp2d._change_data()

        # print(f"Time to update 2D plot: {time.perf_counter() - st}")

    def _update_primary_dim_xhs_from_sliders(self):
        """Update the primary dims crosshairs from the sliders."""
        # Disconnect signal to avoid circular updates
        signal = self.primary_dims_xh.xh.sigPositionChanged
        signal.disconnect(self._update_dim01_slider_position)

        pos = []
        for i in range(2):
            index = self.dim01_sliders[i].value()
            pos.append(self.coarsened_coords[i][index])
        self.primary_dims_xh.set_pos((pos[0], pos[1]))
        self._update_primary_dim_cursor_stats()

        # Reconnect signal
        signal.connect(self._update_dim01_slider_position)

    def _key_press_event(self, event):
        """Handle key press events."""
        if event.key() in [
            QtCore.Qt.Key.Key_Up,
            QtCore.Qt.Key.Key_Down,
            QtCore.Qt.Key.Key_Left,
            QtCore.Qt.Key.Key_Right,
        ]:
            if event.key() == QtCore.Qt.Key.Key_Up:
                self.dim01_sliders[0].setValue(self.dim01_sliders[0].value() + 1)
            elif event.key() == QtCore.Qt.Key.Key_Down:
                self.dim01_sliders[0].setValue(self.dim01_sliders[0].value() - 1)
            elif event.key() == QtCore.Qt.Key.Key_Left:
                self.dim01_sliders[1].setValue(self.dim01_sliders[1].value() - 1)
            elif event.key() == QtCore.Qt.Key.Key_Right:
                self.dim01_sliders[1].setValue(self.dim01_sliders[1].value() + 1)
            pos = []
            for i in range(2):
                index = self.dim01_sliders[i].value()
                pos.append(self.coarsened_coords[i][index])
            self.primary_dims_xh.set_pos((pos[0], pos[1]))
            self._update_primary_dim_cursor_stats()

    def _reset_roi_plots(self):
        """Reset the ROI plots to the default."""
        self._update_dim01_roi_plot(self.data_primary_tot.values, reset_ROI=True)
        self._update_dim23_roi_plot(self.data_secondary_tot.values, resest_ROI=True)

    def _add_ROI_to_disp2d(self, ROI_points=None):
        """Add the ROI to the right plot."""
        if ROI_points is None:
            ROI_points = self._set_ROI_default_pos(2, 3)
        self.roi_dim23 = pg.PolyLineROI(
            positions=ROI_points,
            pen=pg.mkPen((204, 51, 153), width=2.5),
            handlePen=pg.mkPen((204, 51, 153), width=2),
            closed=True,
        )
        self.roi_disp2d.image_plot.addItem(self.roi_dim23)

    def _get_ROI_from_plot(self):
        # Get the ROI points
        if self.select_from_dim01_selector.isChecked():
            pts_scene = self.roi_dim01.getSceneHandlePositions()
            points = [self.roi_dim01.mapSceneToParent(pt[1]) for pt in pts_scene]
            dims = [self.dims[0], self.dims[1]]
        else:
            pts_scene = self.roi_dim23.getSceneHandlePositions()
            points = [self.roi_dim23.mapSceneToParent(pt[1]) for pt in pts_scene]
            dims = [self.dims[2], self.dims[3]]
        valuesA = [i.y() for i in points]
        valuesB = [i.x() for i in points]
        return {dims[0]: valuesA, dims[1]: valuesB}

    def _set_ROI_from_plot(self):
        """Set the ROI on the data and apply to the plots."""
        ROI_dict = self._get_ROI_from_plot()
        new_data_to_plot = self.data.mask_data(ROI_dict).values
        if self.dims[0] in ROI_dict.keys():
            self._update_dim23_roi_plot(new_data_to_plot)
            self._set_ROI_outline(0, ROI_dict)
        else:
            self._update_dim01_roi_plot(new_data_to_plot)
            self._set_ROI_outline(1, ROI_dict)

    def _set_ROI_from_list(self):
        """Apply the ROI to the data and plots based on list selection"""
        selected_items = self.roi_list_box.selectedItems()
        if len(selected_items) == 0:
            return
        elif len(selected_items) == 1:
            ROI_dict_A = self.ROI_store[selected_items[0].text()]
            # Check this ROI compatible with plot
            if set(ROI_dict_A.keys()) != set(self.dims[:2]) and set(
                ROI_dict_A.keys()
            ) != set(self.dims[2:]):
                selected_items[0].setSelected(False)
                QtWidgets.QMessageBox.warning(
                    self,
                    "ROI cannot be displayed",
                    "ROI definition incompatible with current view.",
                    QtWidgets.QMessageBox.StandardButton.Ok,
                    QtWidgets.QMessageBox.StandardButton.Ok,
                )
                return
            new_data_to_plot = self.data.mask_data(ROI_dict_A).values
        elif len(selected_items) == 2:
            ROI_dict_A = self.ROI_store[selected_items[0].text()]
            ROI_dict_B = self.ROI_store[selected_items[1].text()]
            # Check these ROIs compatible with plot
            if (
                set(ROI_dict_A.keys()) != set(self.dims[:2])
                and set(ROI_dict_A.keys()) != set(self.dims[2:])
                or set(ROI_dict_B.keys()) != set(ROI_dict_A.keys())
            ):
                for item in selected_items:
                    item.setSelected(False)
                QtWidgets.QMessageBox.warning(
                    self,
                    "ROI cannot be displayed",
                    "ROI definitions incompatible with each other or current view.",
                    QtWidgets.QMessageBox.StandardButton.Ok,
                    QtWidgets.QMessageBox.StandardButton.Ok,
                )
                return
            new_data_to_plot = (
                self.data.mask_data(ROI_dict_A) - self.data.mask_data(ROI_dict_B)
            ).values
        if self.dims[0] in ROI_dict_A.keys():
            self._update_dim23_roi_plot(new_data_to_plot)
            if len(selected_items) == 1:
                self._set_ROI_outline(0, ROI_dict_A)
            else:
                self._set_ROI_outline(0, ROI_dict_A, ROI_dict_B)
        else:
            self._update_dim01_roi_plot(new_data_to_plot)
            if len(selected_items) == 1:
                self._set_ROI_outline(1, ROI_dict_A)
            else:
                self._set_ROI_outline(1, ROI_dict_A, ROI_dict_B)

    def _set_ROI_outline(self, plot_no, ROI_dict_A, ROI_dict_B=None):
        """Set the ROI outline on the plot when a ROI region is activated."""
        if plot_no == 0:
            for i in self.display_ROIs_01:
                self.primary_dims_roi_plot.removeItem(i)
            self.display_ROIs_01 = []
        else:
            for i in self.display_ROIs_23:
                self.roi_disp2d.image_plot.removeItem(i)
            self.display_ROIs_23 = []

        if ROI_dict_B:
            ROIs = [ROI_dict_A, ROI_dict_B]
        else:
            ROIs = [ROI_dict_A]
        dimA = self.dims[0] if plot_no == 0 else self.dims[2]
        dimB = self.dims[1] if plot_no == 0 else self.dims[3]
        for ROI in ROIs:
            dimA_data = ROI[dimA]
            dimA_data.extend([dimA_data[0]])
            dimB_data = ROI[dimB]
            dimB_data.extend([dimB_data[0]])

            if plot_no == 0:
                line_plot = pg.PlotDataItem(dimB_data, dimA_data, pen="g")
                self.primary_dims_roi_plot.addItem(line_plot)
                self.display_ROIs_01.append(line_plot)
            else:
                line_plot = pg.PlotDataItem(dimB_data, dimA_data, pen="g")
                self.roi_disp2d.image_plot.addItem(line_plot)
                self.display_ROIs_23.append(line_plot)

    def _update_dim01_roi_plot(self, data, reset_ROI=False):
        """Update the primary dims ROI plot.

        Parameters
        ------------
        data : np.ndarray
            The data to display.
        """
        # Reset the data on the plots
        self.primary_dims_roi_plot_image_item.setImage(data)
        # Reset colour scales
        crange = (float(data.min()), float(data.max()))
        self.primary_dims_roi_plot_colorbar.hi_lim = crange[1]
        self.primary_dims_roi_plot_colorbar.lo_lim = crange[0]
        self.primary_dims_roi_plot_colorbar.setLevels(crange)

        if reset_ROI:
            # Reset the ROI on the left plot
            for i in self.display_ROIs_01:
                self.primary_dims_roi_plot.removeItem(i)
            self.display_ROIs_01 = []
            self.roi_dim01.setSize(1)
            self.roi_dim01.setPos([0, 0])
            self.roi_dim01.setAngle(0)
            self.roi_dim01.setPoints(self._set_ROI_default_pos(0, 1))

    def _update_dim23_roi_plot(self, data, resest_ROI=False):
        """Update the secondary dims ROI plot.

        Parameters
        ------------
        data : np.ndarray
            The data to display.
        """
        if not resest_ROI:
            pts_scene = self.roi_dim23.getSceneHandlePositions()
            ROI_points = [self.roi_dim23.mapSceneToParent(pt[1]) for pt in pts_scene]
        self.roi_disp2d.data_arrays[0].data = data
        self.roi_disp2d._change_data()

        # Re-add the ROI to the right plot
        if resest_ROI:
            self._add_ROI_to_disp2d()
        else:
            self._add_ROI_to_disp2d(ROI_points)

    def _add_ROI_to_list(self):
        """Add the current ROI to the list."""

        name = self.ROI_name_box.text()
        # Ensure name unique
        names = [
            self.roi_list_box.item(i).text() for i in range(self.roi_list_box.count())
        ]
        if not name or name in names:
            i = 1
            while f"{name}_({i})" in names:
                i += 1
            name = f"{name}_({i})"
        # Add the name to the list
        self.roi_list_box.addItem(name)
        # Add it to ROI store
        self.ROI_store[name] = self._get_ROI_from_plot()

    def _remove_selected_ROI_from_list(self):
        """Remove the selected ROI(s) from the list."""
        selected_items = self.roi_list_box.selectedItems()
        if not selected_items:
            return

        for item in selected_items:
            del self.ROI_store[item.text()]
            self.roi_list_box.takeItem(self.roi_list_box.row(item))

    def _remove_all_ROIs_from_list(self):
        """Remove all ROIs from the list."""
        items = [self.roi_list_box.item(i) for i in range(self.roi_list_box.count())]

        # Create a confirmation dialog
        reply = QtWidgets.QMessageBox.question(
            self,
            "Confirm Deletion",
            "Are you sure you want to delete all ROI(s)?",
            QtWidgets.QMessageBox.StandardButton.Yes
            | QtWidgets.QMessageBox.StandardButton.No,
            QtWidgets.QMessageBox.StandardButton.No,
        )

        # If the user confirms, proceed with deletion
        if reply == QtWidgets.QMessageBox.StandardButton.Yes:
            for item in items:
                del self.ROI_store[item.text()]
                self.roi_list_box.takeItem(self.roi_list_box.row(item))

    def _select_ROI_from_list(self):
        selected_items = self.roi_list_box.selectedItems()
        modifiers = QtWidgets.QApplication.keyboardModifiers()

        if modifiers == QtCore.Qt.KeyboardModifier.ShiftModifier:
            if len(selected_items) > 2:
                self.roi_list_box.blockSignals(True)
                for item in selected_items[:-2]:
                    item.setSelected(False)
                self.roi_list_box.blockSignals(False)
        else:
            if len(selected_items) > 1:
                self.roi_list_box.blockSignals(True)
                for item in selected_items[:-1]:
                    item.setSelected(False)
                self.roi_list_box.blockSignals(False)

    # ##############################
    # Helper functions
    # ##############################
    def _find_nearest_sorted(self, array, value):
        """Find the index of the nearest point to a given value in an ordered NumPy array.

        Parameters
        ------------
        array : np.ndarray
            The input ordered array.
        value : float
            The value to find the nearest neighbor to.

        Returns
        ------------
        int
            The index of the nearest point to the given value.
        """
        array = np.asarray(array)
        idx = np.searchsorted(array, value, side="left")
        if idx > 0 and (
            idx == len(array)
            or np.abs(value - array[idx - 1]) < np.abs(value - array[idx])
        ):
            return idx - 1
        else:
            return idx

    def _find_nearest_n_indexes(self, array, value, n):
        """
        Find the indexes of the N nearest points to a given value in an ordered NumPy array.

        Parameters
        ----------
        array : np.ndarray
            The input ordered array.
        value : float
            The value to find the nearest neighbors to.
        n : int
            The number of nearest neighbors to find.

        Returns
        -------
        np.ndarray
            The indexes of the nearest N points to the given value.
        """
        idx = np.searchsorted(array, value)
        left = np.clip(idx - n, 0, len(array))
        right = np.clip(idx + n, 0, len(array))
        window = array[left:right]
        sorted_indexes = np.argsort(np.abs(window - value))
        nearest_indexes = np.arange(left, right)[sorted_indexes[:n]]
        start, end = nearest_indexes.min(), nearest_indexes.max() + 1
        return slice(start, end)

    def _set_ROI_default_pos(self, dim0, dim1):
        return [
            [
                np.percentile(self.ranges[dim1], 45),
                np.percentile(self.ranges[dim0], 45),
            ],
            [
                np.percentile(self.ranges[dim1], 55),
                np.percentile(self.ranges[dim0], 45),
            ],
            [
                np.percentile(self.ranges[dim1], 55),
                np.percentile(self.ranges[dim0], 55),
            ],
            [
                np.percentile(self.ranges[dim1], 45),
                np.percentile(self.ranges[dim0], 55),
            ],
        ]



import os

import natsort
import numpy as np
import pint_xarray
import xarray as xr
from scipy import constants
from tqdm.notebook import tqdm

from peaks.core.fileIO.base_arpes_data_classes.base_arpes_data_class import (
    BaseARPESDataLoader,
)
from peaks.core.fileIO.base_data_classes.base_photon_source_classes import (
    BasePumpProbeClass,
)
from peaks.core.fileIO.loc_registry import register_loader
from peaks.core.utils.misc import analysis_warning

ureg = pint_xarray.unit_registry


@register_loader
class ArtemisPhoibos(BaseARPESDataLoader, BasePumpProbeClass):
    """Data loader for Artemis TR-ARPES system, with the Specs Phoibos analyser."""

    _loc_name = "CLF_Artemis"
    _loc_description = "SPECS Phoibos 100 analyser at CLF Artemis facility"
    _loc_url = "https://www.clf.stfc.ac.uk/Pages/Artemis.aspx"
    _analyser_slit_angle = 90 * ureg.deg

    _manipulator_name_conventions = {
        "polar": "polar",
        "azi": "azi",
        "x1": "x",
        "x2": "y",
        "x3": "z",
    }

    _metadata_parsers = [
        "_parse_analyser_metadata",
        "_parse_manipulator_metadata",
        "_parse_photon_metadata",
        "_parse_temperature_metadata",
        "_parse_pump_photon_metadata",
    ]

    _desired_dim_order = [
        "t",
        "x3",
        "x2",
        "x1",
        "polar",
        "tilt",
        "azi",
        "eV",
        "theta_par",
    ]

    _analyser_include_in_metadata_warn = [
        "eV",
        "PE",
        "sweeps",
        "azi",
    ]
    _manipulator_exclude_from_metadata_warn = ["x1", "x2", "x3", "polar", "tilt", "azi"]
    _temperature_exclude_from_metadata_warn = [
        "cryostat",
        "setpoint",
        "shield",
    ]

    @classmethod
    def _load_data(cls, fpath, lazy, **kwargs):
        """Artemis core data loading

        Parameters
        ----------
        fpath : str
            Path to folder containing time series data. In Artemis structure, should contain Info.tsv,
            Stat Log.tsv and N=xxx folders with the data
        lazy : bool
            Not supported for Artemis data
        kwargs : dict
            Additional keyword arguments for the loader. Defaults used if these are not specified:
            - for image transformation:
                - Ang_offset_px
                - Edge_pos
                - Edge_slope
                - calib2d_fpath
            - for partial data loading:
                - num_scans

        """

        if os.path.splitext(fpath)[1] == ".h5":
            # If the file is an HDF5 file, should be a FeSuMa scan; use the BaseFeSuMaDataLoader
            from peaks.core.fileIO.base_arpes_data_classes.base_fesuma_class import (
                BaseFeSuMaDataLoader,
            )

            data_dict = BaseFeSuMaDataLoader._load_data(fpath, lazy, **kwargs)
            if "delay_pos" in data_dict.get("dims") and "t" not in data_dict.get("dims"):
                coords = data_dict.get("coords")
                units = data_dict.get("units")
                delay_pos_m = (
                    coords.get("delay_pos") * ureg(units.get("delay_pos"))
                ).to("m")
                t = ((delay_pos_m - delay_pos_m[0]) * 2 / ureg.c).to("fs").magnitude
                coords["t"] = t
                coords["delay_pos"] = ("t", coords.pop("delay_pos"))
                new_dims = []
                for dim in data_dict.get("dims"):
                    if dim == "delay_pos":
                        new_dims.append("t")
                    else:
                        new_dims.append(dim)
                data_dict["dims"] = new_dims
                data_dict.get("units")["t"] = "fs"

            return data_dict

        # Parse keyword arguments
        Ang_Offset_px = kwargs.pop("Ang_Offset_px", 0)
        Edge_pos = kwargs.pop("Edge_pos", 10)
        Edge_slope = kwargs.pop("Edge_slope", -1)
        num_scans = kwargs.pop("num_scans", None)
        calib2d_fpath = kwargs.pop(
            "calib2d_fpath",
            os.path.join(
                os.path.dirname(__file__), "calib_files", "Artemis_phoibos100.calib2d"
            ),
        )
        if kwargs:
            raise ValueError(
                f"Unexpected keyword arguments: {kwargs}. "
                f"Valid arguments: Ang_Offset_px, Edge_pos, Edge_slope, num_scans, calib2d_fpath"
            )

        # Read folder structure and get test scan
        scan_folders = cls._get_scan_folders(fpath)
        test_scan = cls._get_test_scan(fpath, scan_folders)
        nx_pixel, ny_pixel = test_scan.shape

        # Load and cache metadata from info file
        metadata_dict = cls._load_metadata(fpath, return_dict_with_raw_keys=True)
        metadata_dict["num_scans"] = num_scans  # Add passed num_scans
        cls._metadata_cache[fpath] = metadata_dict

        # Determine some core parameters / set defaults
        binning = int(1920 / nx_pixel)
        PixelSize = 0.00645
        magnification = 4.41
        WF = 4.215
        KE = float(metadata_dict["KE"])
        PE = float(metadata_dict["PE"])
        rr = (KE - WF) / PE  # retarding ratio

        # Extract calibration data
        calib2d_in = cls._read_calib2d_file(calib2d_fpath)
        calib2d, lens_rr_data = cls._extract_calibration_info(calib2d_in, metadata_dict)

        # Interpolate calibration data
        calib2d = cls._interpolate_calibration_data(
            calib2d, lens_rr_data, rr, metadata_dict
        )

        # Calculate transformations
        eV_xarray, theta_par_xarray, E_correction_xarray, Angular_correction_xarray = (
            cls._calculate_transformations(
                calib2d,
                metadata_dict,
                nx_pixel,
                ny_pixel,
                binning,
                magnification,
                PixelSize,
                Ang_Offset_px,
                Edge_pos,
                Edge_slope,
            )
        )

        # Load and transform data
        return cls._load_and_transform_data(
            fpath,
            scan_folders,
            num_scans,
            nx_pixel,
            ny_pixel,
            eV_xarray,
            theta_par_xarray,
            E_correction_xarray,
            Angular_correction_xarray,
            metadata_dict,
            WF,
        )

    @staticmethod
    def _get_scan_folders(fname):
        """Retrieve list of scan folders from the data directory."""
        file_list = natsort.natsorted(os.listdir(fname))
        scan_folders = [item for item in file_list if "N=" in item]
        return scan_folders

    @staticmethod
    def _get_test_scan(fname, scan_folders):
        """Load an example dispersion to determine image size."""
        test_scan_path = os.path.join(
            fname,
            scan_folders[0],
            natsort.natsorted(os.listdir(os.path.join(fname, scan_folders[0])))[0],
        )
        test_scan = np.loadtxt(test_scan_path)
        return test_scan

    @staticmethod
    def _read_calib2d_file(calib2d_fpath):
        """Read the calib2d file containing calibration data."""
        with open(calib2d_fpath, "r") as f:
            calib2d_in = f.readlines()
        return calib2d_in

    @staticmethod
    def _extract_calibration_info(calib2d_in, meta):
        """Extract calibration data relevant to the current lens mode."""
        calib2d = {}
        lens_rr_data = {
            "lens_rr": [],
            "aInner": [],
            "Da1": {"Da1_1": [], "Da1_2": [], "Da1_3": []},
            "Da3": {"Da3_1": [], "Da3_2": [], "Da3_3": []},
            "Da5": {"Da5_1": [], "Da5_2": [], "Da5_3": []},
            "Da7": {"Da7_1": [], "Da7_2": [], "Da7_3": []},
        }

        i0 = None
        i1 = None

        for ct, line in enumerate(calib2d_in):
            # Extract constants
            if "eRange" in line:
                calib2d["eRange"] = [float(j) for j in line.split()[2:4]]
            if "De1" in line:
                calib2d["De1"] = float(line.split("= ")[1])

            # Find relevant block
            if meta["LensMode"] in line and "#" in line:
                i0 = ct + 1
            elif "# ===" in line and i0 and ct - i0 > 2:
                i1 = ct
                break

        calib_lens = calib2d_in[i0:i1] if i1 else calib2d_in[i0:]

        for line in calib_lens:
            # Static parameters
            if "aUnit" in line:
                calib2d["aUnit"] = line.split('"')[1]
            if "aRange" in line:
                calib2d["aRange"] = [float(j) for j in line.split()[2:4]]
            if "eShift" in line:
                calib2d["eShift"] = [float(j) for j in line.split()[2:5]]

            # Retarding ratio-dependent parameters
            if meta["LensMode"] in line and "@" in line:
                rr_value = float(line.split("@")[1].split("]")[0])
                lens_rr_data["lens_rr"].append(rr_value)
            if "aInner" in line:
                aInner_value = float(line.split("= ")[1].split("#")[0])
                lens_rr_data["aInner"].append(aInner_value)
            for Da in ["Da1", "Da3", "Da5", "Da7"]:
                if Da in line:
                    temp_values = [float(j) for j in line.split()[2:5]]
                    lens_rr_data[Da][f"{Da}_1"].append(temp_values[0])
                    lens_rr_data[Da][f"{Da}_2"].append(temp_values[1])
                    lens_rr_data[Da][f"{Da}_3"].append(temp_values[2])

        return calib2d, lens_rr_data

    @staticmethod
    def _interpolate_calibration_data(calib2d, lens_rr_data, rr, meta):
        """Interpolate calibration data for the current retarding ratio."""
        lens_rr = lens_rr_data["lens_rr"]
        calib2d["aInner"] = np.interp(rr, lens_rr, lens_rr_data["aInner"])
        eV_coords = np.array(calib2d["eShift"]) * float(meta["PE"]) + float(meta["KE"])

        for Da in ["Da1", "Da3", "Da5", "Da7"]:
            Da_values = np.array(
                [
                    np.interp(rr, lens_rr, lens_rr_data[Da][f"{Da}_1"]),
                    np.interp(rr, lens_rr, lens_rr_data[Da][f"{Da}_2"]),
                    np.interp(rr, lens_rr, lens_rr_data[Da][f"{Da}_3"]),
                ]
            )
            calib2d[Da] = xr.DataArray(
                data=Da_values, dims="eV", coords={"eV": eV_coords}
            )

        # Polynomial fits for Da coefficients
        calib2d["Da_value"] = eV_coords
        for Da in ["Da1", "Da3", "Da5", "Da7"]:
            calib2d[f"{Da}_coeff"] = np.polyfit(calib2d["Da_value"], calib2d[Da], 2)
        return calib2d

    @staticmethod
    def _calculate_transformations(
        calib2d,
        meta,
        nx_pixel,
        ny_pixel,
        binning,
        magnification,
        PixelSize,
        Ang_Offset_px,
        Edge_pos,
        Edge_slope,
    ):
        """Calculate energy and angular correction transformations."""
        # Energy range
        Ek_low = float(meta["KE"]) + float(calib2d["eRange"][0]) * float(meta["PE"])
        Ek_high = float(meta["KE"]) + float(calib2d["eRange"][1]) * float(meta["PE"])
        theta_par_low = float(calib2d["aRange"][0])
        theta_par_high = float(calib2d["aRange"][1])

        # Create energy and angle arrays
        eV_xarray = xr.DataArray(
            data=np.linspace(Ek_low, Ek_high, nx_pixel),
            dims="eV",
            coords={"eV": np.linspace(Ek_low, Ek_high, nx_pixel)},
        )
        theta_par_xarray = xr.DataArray(
            data=np.linspace(theta_par_low, theta_par_high, ny_pixel),
            dims="theta_par",
            coords={"theta_par": np.linspace(theta_par_low, theta_par_high, ny_pixel)},
        )

        # Energy correction
        E_correction = np.round(
            (eV_xarray - float(meta["KE"]))
            / float(meta["PE"])
            / float(calib2d["De1"])
            / magnification
            / (PixelSize * binning)
            + nx_pixel / 2
        )

        # Interpolation for energy correction
        eV_i = xr.DataArray(
            data=eV_xarray.data,
            dims="eV_i",
            coords={"eV_i": np.linspace(0, nx_pixel - 1, nx_pixel)},
        )
        E_correction_interp = xr.DataArray(
            data=E_correction,
            dims="eV_i",
            coords={"eV_i": np.linspace(0, nx_pixel - 1, nx_pixel)},
        )
        E_correction_xarray = xr.DataArray(
            data=eV_i.interp(eV_i=E_correction_interp).data,
            dims="eV",
            coords={"eV": eV_xarray.eV},
        )

        # Interpolate Da coefficients
        Da_coeffs = {}
        for Da in ["Da1", "Da3", "Da5", "Da7"]:
            Da_coeffs[Da] = np.polyval(calib2d[f"{Da}_coeff"], eV_xarray)
            Da_coeffs[Da] = xr.DataArray(
                data=Da_coeffs[Da], dims="eV", coords={"eV": eV_xarray.eV}
            )

        # Calculate zInner_data and zInner_Diff_data
        theta_par = theta_par_xarray
        zInner_data = (
            Da_coeffs["Da1"] * theta_par
            + 1e-2 * Da_coeffs["Da3"] * theta_par**3
            + 1e-4 * Da_coeffs["Da5"] * theta_par**5
            + 1e-6 * Da_coeffs["Da7"] * theta_par**7
        )
        zInner_Diff_data = (
            Da_coeffs["Da1"]
            + 3e-2 * Da_coeffs["Da3"] * theta_par**2
            + 5e-4 * Da_coeffs["Da5"] * theta_par**4
            + 7e-6 * Da_coeffs["Da7"] * theta_par**6
        )

        # MCP Position calculations
        aInner = calib2d["aInner"]
        MCP_Position_mm_1 = zInner_data.where(abs(theta_par) <= aInner).fillna(0)
        MCP_Position_mm_2 = (
            (
                np.sign(theta_par)
                * (zInner_data + (abs(theta_par) - aInner) * zInner_Diff_data)
            )
            .where(abs(theta_par) > aInner)
            .fillna(0)
        )
        MCP_Position_mm = MCP_Position_mm_1 + MCP_Position_mm_2

        # Edge correction
        Edge_Coef = (
            np.tan(np.radians(Edge_slope))
            / MCP_Position_mm.interp(eV=float(meta["KE"]), theta_par=Edge_pos).data
            / float(meta["PE"])
            / float(calib2d["De1"])
        )
        w_LinearCorrection = 1 / (1 + (Edge_Coef * (eV_xarray - float(meta["KE"]))))

        Angular_correction = (
            (w_LinearCorrection * MCP_Position_mm)
            / magnification
            / (PixelSize * binning)
            + ny_pixel / 2
            + Ang_Offset_px
        )

        # Interpolation for angular correction
        theta_par_i = xr.DataArray(
            data=theta_par_xarray.data,
            dims="theta_par_i",
            coords={"theta_par_i": np.linspace(0, ny_pixel - 1, ny_pixel)},
        )
        Angular_correction_interp = xr.DataArray(
            data=Angular_correction,
            dims=["eV_i", "theta_par_i"],
            coords={
                "eV_i": np.linspace(0, nx_pixel - 1, nx_pixel),
                "theta_par_i": np.linspace(0, ny_pixel - 1, ny_pixel),
            },
        )
        Angular_correction_xarray = xr.DataArray(
            data=theta_par_i.interp(theta_par_i=Angular_correction_interp).data,
            dims=["eV", "theta_par"],
            coords={"eV": eV_xarray.eV, "theta_par": theta_par_xarray.theta_par},
        )

        return (
            eV_xarray,
            theta_par_xarray,
            E_correction_xarray,
            Angular_correction_xarray,
        )

    @staticmethod
    def _load_and_transform_data(
        fname,
        scan_folders,
        num_scans,
        nx_pixel,
        ny_pixel,
        eV_xarray,
        theta_par_xarray,
        E_correction_xarray,
        Angular_correction_xarray,
        meta,
        WF,
    ):
        """Load raw data and apply transformations."""
        # Determine scan folder
        if num_scans:
            scan_folder = os.path.join(fname, f"N={num_scans}")
        else:
            scan_folder = os.path.join(fname, scan_folders[-1])
            num_scans = int(scan_folders[-1].split("=")[1])

        # Get scan names
        scan_names = [
            i for i in natsort.natsorted(os.listdir(scan_folder)) if ".tsv" in i
        ]
        num_delays = len(scan_names)

        # Initialize data arrays
        delay_pos = np.empty(num_delays)
        data = np.empty((nx_pixel, ny_pixel, num_delays))

        # Load data
        for i, scan_name in tqdm(
            enumerate(scan_names), total=len(scan_names), desc="Loading scans"
        ):
            delay_pos[i] = float(scan_name.split(".tsv")[0])
            data[:, :, i] = np.loadtxt(os.path.join(scan_folder, scan_name))

        # Calculate delay times
        delay_time = (
            (delay_pos - float(meta["Time Zero"])) * 2 / constants.c * 1e12
        )  # Convert to fs

        # Create xarray DataArray
        data_xarray = xr.DataArray(
            data=data,
            dims=["eV", "theta_par", "t"],
            coords={
                "eV": eV_xarray.eV,
                "theta_par": theta_par_xarray.theta_par,
                "t": delay_time,
            },
        )

        # Apply transformations
        spectrum = data_xarray.interp(
            eV=E_correction_xarray, theta_par=Angular_correction_xarray
        )

        return {
            "spectrum": spectrum.data,
            "dims": ["eV", "theta_par", "t"],
            "coords": {
                "eV": eV_xarray.eV.data - WF,
                "theta_par": theta_par_xarray.theta_par.data,
                "t": delay_time,
                "delay_pos": ("t", delay_pos),
            },
            "units": {
                "eV": "eV",
                "theta_par": "deg",
                "t": "fs",
                "delay_pos": "mm",
                "spectrum": "counts",
            },
        }

    @classmethod
    def _load_metadata(cls, fpath, return_dict_with_raw_keys=False):
        if os.path.splitext(fpath)[1] == ".h5":
            # If the file is an HDF5 file, should be a FeSuMa scan; use the BaseFeSuMaDataLoader
            from peaks.core.fileIO.base_arpes_data_classes.base_fesuma_class import (
                BaseFeSuMaDataLoader,
            )

            return BaseFeSuMaDataLoader._load_metadata(fpath)

        # Check if there is a cached version (only cache in this loader with keys in artemis format)
        metadata_dict_artemis_keys = cls._metadata_cache.get(fpath)

        if not metadata_dict_artemis_keys:
            # Load metadata from info file
            with open(os.path.join(fpath, "Info.tsv"), "r") as f:
                meta = f.readlines()
            metadata_dict_artemis_keys = {
                item.split(":", 1)[0]: item.split(":", 1)[1]
                for item in meta[0].split(";")
                if len(item.split(":")) == 2
            }
            for key, value in metadata_dict_artemis_keys.copy().items():
                try:
                    if key != "Delay List":
                        metadata_dict_artemis_keys[key] = float(value)
                except ValueError:
                    pass

            # Load stats file for parsing e.g. sample temperature
            metadata_dict_artemis_keys.update(cls._load_stats(fpath))

        if return_dict_with_raw_keys:
            return metadata_dict_artemis_keys

        # Otherwise parse to peaks format
        return cls._parse_metadata(metadata_dict_artemis_keys, fpath)

    @staticmethod
    def _load_stats(fpath):
        # Load and parse stats file
        stat_log = np.loadtxt(os.path.join(fpath, "Stat Log.tsv"))
        return {
            "measurement_time": stat_log[:, 0],
            "measurement_stats": stat_log[:, 1],
            "sample_temp": stat_log[:, 2],
        }

    @classmethod
    def _get_max_num_scans(cls, meta, fpath):
        if meta.get("num_scans"):
            return int(meta["num_scans"])
        else:
            scan_folders = cls._get_scan_folders(fpath)
            return max(int(folder.split("=")[1]) for folder in scan_folders)

    @classmethod
    def _parse_metadata(cls, metadata_dict_artemis_keys, fpath):
        """Convert Artemis metadata dictionary keys to peaks keys, and add some default metadata."""

        def _parse_delays(input_string):
            if input_string is None:
                return None
            ranges = (
                [input_string] if "LF" not in input_string else input_string.split("#LF")
            )
            delays = []

            for r in ranges:
                if "_" in r:
                    start, stop, step = map(int, r.split("_"))
                    delays.extend(
                        range(start, stop + step, step)
                    )  # +step to include the endpoint if it's on the step

            delays = np.asarray(delays)
            return np.array([delays[0], delays[-1]]) * ureg("fs")

        def _parse_tempterature(input_array):
            if input_array is None:
                return None
            input_array = input_array * ureg("K")
            # Check if temperature varies significantly over the run
            temp_diff = abs(np.ptp(input_array))
            if temp_diff > 5 * ureg("K"):
                analysis_warning(
                    f"Significant temperature variation of {temp_diff:.2f} over the run, from "
                    f"{min(input_array):.2f} to {max(input_array):.2f}. Average value of {np.mean(input_array):.2f} "
                    f"used in metadata. To view the full data, load the stats file: "
                    f"`from peaks.core.fileIO.loaders.clf import ArtemisPhoibos; ArtemisPhoibos.load_stats(fpath)`",
                    "warning",
                    "Temperature variation warning",
                )
            return np.mean(input_array)

        # Define metadata
        return {
            "analyser_model": "Phoibos 100",
            "analyser_slit_width": None,
            "analyser_slit_width_identifier": None,
            "analyser_eV": (
                metadata_dict_artemis_keys.get("KE") * ureg("eV")
                if metadata_dict_artemis_keys.get("KE")
                else None
            ),
            "analyser_eV_type": "kinetic",
            "analyser_step_size": None,
            "analyser_PE": (
                metadata_dict_artemis_keys.get("PE") * ureg("eV")
                if metadata_dict_artemis_keys.get("PE")
                else None
            ),
            "analyser_sweeps": cls._get_max_num_scans(metadata_dict_artemis_keys, fpath),
            "analyser_dwell": None,
            "analyser_lens_mode": metadata_dict_artemis_keys.get("LensMode"),
            "analyser_acquisition_mode": "fixed",
            "analyser_polar": None,
            "analyser_tilt": None,
            "analyser_azi": cls._analyser_slit_angle,
            "timestamp": metadata_dict_artemis_keys.get("Run Start"),
            "manipulator_polar": None,
            "manipulator_tilt": None,
            "manipulator_azi": None,
            "manipulator_x1": None,
            "manipulator_x2": None,
            "manipulator_x3": None,
            "photon_hv": "Excitation Energy",
            "pump_hv": None,
            "pump_power": None,
            "pump_delay": _parse_delays(metadata_dict_artemis_keys.get("Delay List")),
            "pump_polarisation": None,
            "pump_t0_position": float(metadata_dict_artemis_keys.get("Time Zero"))
            * ureg("mm"),
            "temperature_sample": _parse_tempterature(
                metadata_dict_artemis_keys.get("sample_temp")
            ),
        }

    @classmethod
    def load_stats(cls, fpath, metadata=True, quiet=False):
        """Return the signal and temperature stats vs. time from an Artemis measurement run

        Parameters
        ----------
        fpath : str
            Path to the folder containing the Artemis data

        metadata : bool
            Whether to include additional metadata in the returned dataset

        quiet : bool
            Whether to suppress warnings, defaults to False

        Returns
        -------
        xarray.Dataset
            Dataset containing the temperature and signal stats vs. measurement time
        """

        stats = cls._load_stats(fpath)
        stats_da = xr.DataArray(
            data=stats["measurement_stats"],
            dims="measurement_time",
            coords={"measurement_time": stats["measurement_time"]},
            name="stats",
        ).pint.quantify({"stats": "counts", "measurement_time": "min"})
        temp_da = xr.DataArray(
            data=stats["sample_temp"],
            dims="measurement_time",
            coords={"measurement_time": stats["measurement_time"]},
            name="temperature",
        ).pint.quantify({"temperature": "K", "measurement_time": "min"})
        ds = xr.Dataset({"stats": stats_da, "temperature": temp_da})

        # Load the metadata for adding to the ds if desired
        metadata_dict = cls.load_metadata(
            fpath=fpath,
            loc=cls._loc_name,
            quiet=quiet,
            load_metadata_from_file=metadata,
        )

        ds.attrs.update(metadata_dict)

        return ds



import os
import shutil
import tempfile
import time
import zipfile
from copy import deepcopy
from pathlib import Path

import matplotlib.pyplot as plt
import requests
from requests.adapters import HTTPAdapter
from tqdm.notebook import tqdm
from urllib3.util.retry import Retry

from peaks.core.fileIO.data_loading import load
from peaks.core.utils.misc import analysis_warning

ROOT_URL = "https://zenodo.org/api/records/15928652/files"


class ZenodoDownloader:
    """Helper class to download files from Zenodo.

    If the data is not public, a Zenodo token must be provided. This will be taken
    from the environment variable `ZENODO_TOKEN` if not explicitly provided.

    If `LOCAL_MIRROR_PATH` is set, files will be copied from this local directory
    with Zenodo serving as the fallback (used in CI)."""

    def __init__(self, file_list, token=None):
        self.root_url = ROOT_URL.rstrip("/")
        self.file_list = file_list
        self.token = token or os.getenv("ZENODO_TOKEN")
        self.local_mirror = os.getenv("LOCAL_MIRROR_PATH")
        self._tempdir_context = None
        self.downloaded_files = {}

    def _get_local_path(self, filename):
        """Check if local mirror available and if files exist there."""
        if self.local_mirror and os.path.isdir(self.local_mirror):
            path = os.path.join(self.local_mirror, filename)
            if os.path.isfile(path):
                return path
        return None

    def _make_headers_if_needed(self, url):
        if "draft" in url and self.token:
            return {"Authorization": f"Bearer {self.token}"}
        return {}

    def _download_with_progress(self, url, dest_path):
        headers = self._make_headers_if_needed(url)

        session = requests.Session()
        session.headers.update(
            {"User-Agent": "peaks (research software; https://github.com/phrgab/peaks)"}
        )
        retries = Retry(
            total=5,
            backoff_factor=1,
            status_forcelist=[429, 502, 503, 504],
            allowed_methods=["GET"],
            raise_on_status=False,
        )
        adapter = HTTPAdapter(max_retries=retries)
        session.mount("http://", adapter)
        session.mount("https://", adapter)

        proxies = {"http": os.getenv("http_proxy"), "https": os.getenv("https_proxy")}

        try:
            response = session.get(
                url,
                headers=headers,
                proxies=proxies,
                stream=True,
                timeout=(10, 300),  # (connect timeout, read timeout)
            )
            response.raise_for_status()
        except requests.exceptions.RequestException as e:
            raise RuntimeError(f"Failed to download {url}: {e}") from e

        total = int(response.headers.get("content-length", 0))

        with (
            open(dest_path, "wb") as f,
            tqdm(
                desc=os.path.basename(dest_path),
                total=total,
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
                leave=False,
            ) as bar,
        ):
            for chunk in response.iter_content(chunk_size=1024 * 1024):
                if chunk:  # filter out keep-alive chunks
                    f.write(chunk)
                    f.flush()
                    bar.update(len(chunk))

    def download(self):
        if self._tempdir_context is not None:
            return self.downloaded_files

        self._tempdir_context = tempfile.TemporaryDirectory()
        tmpdir = self._tempdir_context.name

        with tqdm(
            total=len(self.file_list), desc="Downloading sample file(s)", unit="file"
        ) as overall_bar:
            for filename in self.file_list:
                local_path = self._get_local_path(filename)
                dest_path = os.path.join(tmpdir, filename)
                if local_path:
                    shutil.copy(local_path, dest_path)
                else:
                    url = f"{self.root_url}/{filename}/content"
                    self._download_with_progress(url, dest_path)
                    time.sleep(1)  # pause between Zenodo API downloads for fair use
                self.downloaded_files[filename] = dest_path
                overall_bar.update(1)

        return self.downloaded_files

    @property
    def fpath(self):
        """Path to the temporary directory containing downloaded files."""
        if not self.downloaded_files:
            raise ValueError(
                "No files have been downloaded yet or the temporary files "
                "have been deleted. Call `download()` to download the files."
            )
        return self._tempdir_context.name

    def cleanup(self, quiet=False):
        if self._tempdir_context is not None:
            self._tempdir_context.cleanup()
            self._tempdir_context = None
            self.downloaded_files = {}
            analysis_warning(
                "Temporary files have been deleted. You can now safely close the notebook.",
                title="Cleanup Complete",
                warn_type="success",
                quiet=quiet,
            )
        else:
            analysis_warning(
                "No temporary files to clean up. The cleanup method has already been "
                "called or no files were downloaded.",
                title="No Cleanup Needed",
                warn_type="info",
            )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.cleanup(quiet=True)


def get_tutorial1_data():
    """Download the tutorial 1 data from Zenodo."""
    file_list = ["i05-59818.nxs", "i05-59819.nxs", "i05-59853.nxs"]
    downloader = ZenodoDownloader(file_list)
    analysis_warning(
        "Downloading data files for the Tutorial to a temporary directory. The file path"
        " is available via the .fpath attribute. Make sure to call the `cleanup()` "
        "method at the end of the tutorial to delete the temporary files.",
        title="Sample Data Download",
        warn_type="warning",
    )
    downloader.download()

    return downloader


class ExampleData:
    """Class to access example data files for the Peaks package. The data is
    downloaded from Zenodo and cached for subsequent access.
    The files are available as class methods, e.g., `ExampleData.dispersion()`."""

    _cache = {}

    @classmethod
    def _get_and_load(cls, fname, **kwargs):
        """Load data from a file, downloading it if not already cached.
        kwargs are passed to the `load` function."""
        data = cls._cache.get(fname)
        if data is None:
            with ZenodoDownloader([fname]) as downloader:
                downloader.download()
                fpath = os.path.join(downloader.fpath, fname)
                data = load(fpath, **kwargs)
                cls._cache[fname] = data
        return data.copy(deep=True)  # Return a copy to avoid modifying the cached data

    @classmethod
    def _get_and_load_from_zip(cls, fname, **kwargs):
        """Load data from a file inside a zip archive, downloading it if not already
        cached. The zip file is downloaded, extracted, and the specified file is loaded.
        kwargs are passed to the `load` function."""
        data = cls._cache.get(fname)
        if data is None:
            with ZenodoDownloader([fname]) as downloader:
                downloader.download()
                fpath = os.path.join(downloader.fpath, fname)
                # unzip the file
                with zipfile.ZipFile(fpath, "r") as zip_ref:
                    zip_ref.extractall(downloader.fpath)
                # Load the data from the extracted file
                fpath = os.path.join(downloader.fpath, os.path.splitext(fname)[0])
                data = load(fpath, **kwargs)
                cls._cache[fname] = data
        return data.copy(deep=True)  # Return a copy to avoid modifying the cached data

    @classmethod
    def dispersion(cls):
        return cls._get_and_load("i05-59819.nxs")

    @classmethod
    def dispersion2a(cls):
        return cls._get_and_load("210326_GM2-667_GK_1.xy")

    @classmethod
    def dispersion2b(cls):
        return cls._get_and_load("210326_GM2-667_GK_2.xy")

    @classmethod
    def dispersion2c(cls):
        return cls._get_and_load("210326_GM2-667_GK_3.xy")

    @classmethod
    def dispersion3(cls):
        return cls._get_and_load("i05-1-34301.nxs")

    @classmethod
    def gold_reference(cls):
        return cls._get_and_load("i05-59853.nxs")

    @classmethod
    def gold_reference2(cls):
        return cls._get_and_load("Gold.xy")

    @classmethod
    def gold_reference3(cls):
        return cls._get_and_load("i05-70214.nxs")

    @classmethod
    def gold_reference4(cls):
        return cls._get_and_load("Ep20eV.xy")

    @classmethod
    def FS(cls):
        return cls._get_and_load("i05-59818.nxs")

    @classmethod
    def hv_map(cls):
        return cls._get_and_load("i05-69294.nxs")

    @classmethod
    def SM(cls):
        return cls._get_and_load("i05-1-24270_sm.nc")

    @classmethod
    def nano_focus(cls):
        return cls._get_and_load("i05-1-49292.nxs")

    @classmethod
    def nano_focus_w_I0norm(cls):
        return cls._get_and_load("i05-1-49292.nxs", norm_by_I0=True)

    @classmethod
    def tr_arpes(cls):
        # Need to download and unzip the file to simulate the original data structure
        return cls._get_and_load_from_zip("029 Gr.zip")

    @classmethod
    def tr_arpes2(cls):
        return cls._get_and_load_from_zip("028 Gr.zip")

    @classmethod
    def xps(cls):
        return cls._get_and_load("i05-1-49260.nxs")

    @classmethod
    def structure(cls):
        """If `LOCAL_MIRROR_PATH` is set, files will be copied from this local directory
        with the database sever as the fallback (used in CI)."""
        data = cls._cache.get("structure")
        if data is None:
            local_mirror = os.getenv("LOCAL_MIRROR_PATH")
            if local_mirror:
                local_path = os.path.join(local_mirror, "4515175.cif")
                if os.path.exists(local_path):
                    data = load(local_path)
            if data is None:
                url = "https://qiserver.ugr.es/cod/4515175.cif"
                response = requests.get(url)
                if response.status_code == 200:
                    with tempfile.NamedTemporaryFile(
                        "w+", suffix=".cif", delete=True
                    ) as tmp:
                        tmp.write(response.text)
                        tmp.flush()
                        data = load(tmp.name)
                else:
                    raise ValueError(
                        f"Failed to download CIF. HTTP status code: {response.status_code}"
                    )
            cls._cache["structure"] = data
        return deepcopy(data)

    @classmethod
    def cleanup(cls):
        cls._cache.clear()


def plot_tutorial_example_figure(fname, figsize=(16, 8)):
    """Plot an example figure in the tutorial notebooks."""
    tutorial_dir = os.path.join(
        Path(__file__).resolve().parent.parent.parent.parent, "tutorials", "figs"
    )
    fig_path = os.path.join(tutorial_dir, fname)
    plt.figure(figsize=figsize)
    plt.imshow(plt.imread(fig_path))
    plt.axis("off")
    plt.show()
